from datetime import datetime
from time import sleep
from unittest import mock

import pytest

from cognite.client import utils


class TestDatetimeToMs:
    def test_datetime_to_ms(self):
        from datetime import datetime

        assert utils._time.datetime_to_ms(datetime(2018, 1, 31)) == 1517356800000
        assert utils._time.datetime_to_ms(datetime(2018, 1, 31, 11, 11, 11)) == 1517397071000
        assert utils._time.datetime_to_ms(datetime(100, 1, 31)) == -59008867200000
        with pytest.raises(TypeError):
            utils._time.datetime_to_ms(None)

    def test_ms_to_datetime(self):
        from datetime import datetime

        assert utils._time.ms_to_datetime(1517356800000) == datetime(2018, 1, 31)
        assert utils._time.ms_to_datetime(1517397071000) == datetime(2018, 1, 31, 11, 11, 11)
        with pytest.raises(ValueError, match="greater than"):
            utils._time.ms_to_datetime(-59008867200000)
        with pytest.raises(TypeError):
            utils._time.ms_to_datetime(None)


class TestTimestampToMs:
    @pytest.mark.parametrize("t", [None, [], {}])
    def test_invalid_type(self, t):
        with pytest.raises(TypeError, match="must be"):
            utils._time.timestamp_to_ms(t)

    def test_ms(self):
        assert 1514760000000 == utils._time.timestamp_to_ms(1514760000000)
        assert 1514764800000 == utils._time.timestamp_to_ms(1514764800000)

    def test_datetime(self):
        assert 1514764800000 == utils._time.timestamp_to_ms(datetime(2018, 1, 1))
        assert 1546300800000 == utils._time.timestamp_to_ms(datetime(2019, 1, 1))

    def test_float(self):
        assert 1514760000000 == utils._time.timestamp_to_ms(1514760000000.0)
        assert 1514764800000 == utils._time.timestamp_to_ms(1514764800000.0)

    @mock.patch("cognite.client.utils._time.time.time")
    @pytest.mark.parametrize(
        "time_ago_string, expected_timestamp",
        [
            ("now", 10 ** 12),
            ("1s-ago", 10 ** 12 - 1 * 1000),
            ("13s-ago", 10 ** 12 - 13 * 1000),
            ("1m-ago", 10 ** 12 - 1 * 60 * 1000),
            ("13m-ago", 10 ** 12 - 13 * 60 * 1000),
            ("1h-ago", 10 ** 12 - 1 * 60 * 60 * 1000),
            ("13h-ago", 10 ** 12 - 13 * 60 * 60 * 1000),
            ("1d-ago", 10 ** 12 - 1 * 24 * 60 * 60 * 1000),
            ("13d-ago", 10 ** 12 - 13 * 24 * 60 * 60 * 1000),
            ("1w-ago", 10 ** 12 - 1 * 7 * 24 * 60 * 60 * 1000),
            ("13w-ago", 10 ** 12 - 13 * 7 * 24 * 60 * 60 * 1000),
        ],
    )
    def test_time_ago(self, time_mock, time_ago_string, expected_timestamp):
        time_mock.return_value = 10 ** 9

        assert utils._time.timestamp_to_ms(time_ago_string) == expected_timestamp

    @pytest.mark.parametrize("time_ago_string", ["1s", "4h", "13m-ag", "13m ago", "bla"])
    def test_invalid(self, time_ago_string):
        with pytest.raises(ValueError, match=time_ago_string):
            utils._time.timestamp_to_ms(time_ago_string)

    def test_time_ago_real_time(self):
        expected_time_now = datetime.now().timestamp() * 1000
        time_now = utils._time.timestamp_to_ms("now")
        assert abs(expected_time_now - time_now) < 10

        sleep(0.2)

        time_now = utils._time.timestamp_to_ms("now")
        assert abs(expected_time_now - time_now) > 190

    @pytest.mark.parametrize("t", [-1, datetime(1969, 12, 31), "100000000w-ago"])
    def test_negative(self, t):
        with pytest.raises(ValueError, match="negative"):
            utils._time.timestamp_to_ms(t)


class TestGranularityToMs:
    @pytest.mark.parametrize(
        "granularity, expected_ms",
        [
            ("1s", 1 * 1000),
            ("13s", 13 * 1000),
            ("1m", 1 * 60 * 1000),
            ("13m", 13 * 60 * 1000),
            ("1h", 1 * 60 * 60 * 1000),
            ("13h", 13 * 60 * 60 * 1000),
            ("1d", 1 * 24 * 60 * 60 * 1000),
            ("13d", 13 * 24 * 60 * 60 * 1000),
        ],
    )
    def test_to_ms(self, granularity, expected_ms):
        assert utils._time.granularity_to_ms(granularity) == expected_ms

    @pytest.mark.parametrize("granularity", ["2w", "-3h", "13m-ago", "13", "bla"])
    def test_to_ms_invalid(self, granularity):
        with pytest.raises(ValueError, match=granularity):
            utils._time.granularity_to_ms(granularity)


class TestGranularityUnitToMs:
    @pytest.mark.parametrize(
        "granularity, expected_ms",
        [
            ("1s", 1 * 1000),
            ("13s", 1 * 1000),
            ("1m", 1 * 60 * 1000),
            ("13m", 1 * 60 * 1000),
            ("1h", 1 * 60 * 60 * 1000),
            ("13h", 1 * 60 * 60 * 1000),
            ("1d", 1 * 24 * 60 * 60 * 1000),
            ("13d", 1 * 24 * 60 * 60 * 1000),
        ],
    )
    def test_to_ms(self, granularity, expected_ms):
        assert utils._time.granularity_unit_to_ms(granularity) == expected_ms

    @pytest.mark.parametrize("granularity", ["2w", "-3h", "13m-ago", "13", "bla"])
    def test_to_ms_invalid(self, granularity):
        with pytest.raises(ValueError, match="format"):
            utils._time.granularity_unit_to_ms(granularity)


class TestObjectTimeConversion:
    @pytest.mark.parametrize(
        "item, expected_output",
        [
            ({"created_time": 0}, {"created_time": "1970-01-01 00:00:00"}),
            ({"last_updated_time": 0}, {"last_updated_time": "1970-01-01 00:00:00"}),
            ({"start_time": 0}, {"start_time": "1970-01-01 00:00:00"}),
            ({"end_time": 0}, {"end_time": "1970-01-01 00:00:00"}),
            ({"not_a_time": 0}, {"not_a_time": 0}),
            ([{"created_time": 0}], [{"created_time": "1970-01-01 00:00:00"}]),
            ([{"last_updated_time": 0}], [{"last_updated_time": "1970-01-01 00:00:00"}]),
            ([{"start_time": 0}], [{"start_time": "1970-01-01 00:00:00"}]),
            ([{"end_time": 0}], [{"end_time": "1970-01-01 00:00:00"}]),
            ([{"source_created_time": 0}], [{"source_created_time": "1970-01-01 00:00:00"}]),
            ([{"source_modified_time": 0}], [{"source_modified_time": "1970-01-01 00:00:00"}]),
            ([{"not_a_time": 0}], [{"not_a_time": 0}]),
            ([{"created_time": int(1e15)}], [{"created_time": int(1e15)}]),
        ],
    )
    def test_convert_time_attributes_to_datetime(self, item, expected_output):
        assert expected_output == utils._time.convert_time_attributes_to_datetime(item)



import cProfile
import functools
import gzip
import json
import os
from contextlib import contextmanager

BASE_URL = "https://greenfield.cognitedata.com"


def jsgz_load(s):
    return json.loads(gzip.decompress(s).decode())


@contextmanager
def profilectx():
    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    pr.print_stats(sort="cumtime")


def profile(method):
    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        with profilectx():
            method(*args, **kwargs)

    return wrapper


@contextmanager
def set_request_limit(client, limit):
    limits = [
        "_CREATE_LIMIT",
        "_LIST_LIMIT",
        "_RETRIEVE_LIMIT",
        "_UPDATE_LIMIT",
        "_DELETE_LIMIT",
        "_DPS_LIMIT",
        "_DPS_LIMIT_AGG",
        "_POST_DPS_OBJECTS_LIMIT",
        "_RETRIEVE_LATEST_LIMIT",
    ]

    tmp = {l: 0 for l in limits}
    for limit_name in limits:
        if hasattr(client, limit_name):
            tmp[limit_name] = getattr(client, limit_name)
            setattr(client, limit_name, limit)
    yield
    for limit_name, limit_val in tmp.items():
        if hasattr(client, limit_name):
            setattr(client, limit_name, limit_val)


@contextmanager
def unset_env_var(*name: str):
    if isinstance(name, str):
        name = [name]
    tmp = {}
    for n in name:
        tmp[n] = os.getenv(n)
        if tmp[n] is not None:
            del os.environ[n]
    yield
    for n in name:
        if tmp[n] is not None:
            os.environ[n] = tmp[n]


@contextmanager
def set_env_var(name: str, value: str):
    tmp = os.getenv(name)
    os.environ[name] = value
    yield
    if tmp is not None:
        os.environ[name] = tmp
    else:
        del os.environ[name]



from typing import List, Union

from cognite.client._api_client import APIClient
from cognite.client.data_classes.templates import *


class TemplatesAPI(APIClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.groups = TemplateGroupsAPI(*args, **kwargs)
        self.versions = TemplateGroupVersionsAPI(*args, **kwargs)
        self.instances = TemplateInstancesAPI(*args, **kwargs)
        self.views = TemplateViewsAPI(*args, **kwargs)

    def graphql_query(self, external_id: str, version: int, query: str) -> GraphQlResponse:
        """
        `Run a GraphQL Query.`
        To learn more, see https://graphql.org/learn/

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group to run the query on.
            query (str): The GraphQL query to run.

        Returns:
            GraphQlResponse: the result of the query.

        Examples:
            Run a GraphQL query:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> query = '''
                >>>    {
                >>>        countryQuery {
                >>>           name,
                >>>           demographics {
                >>>               populationSize,
                >>>               growthRate
                >>>           },
                >>>           deaths {
                >>>               datapoints(limit: 100) {
                >>>                   timestamp,
                >>>                   value
                >>>               }
                >>>           }
                >>>        }
                >>>    }
                >>>    '''
                >>> result = c.templates.graphql_query("template-group-ext-id", 1, query)
        """
        path = "/templategroups/{}/versions/{}/graphql"
        path = utils._auxiliary.interpolate_and_url_encode(path, external_id, version)
        response = self._post(path, {"query": query})
        return GraphQlResponse._load(response.json())


class TemplateGroupsAPI(APIClient):
    _RESOURCE_PATH = "/templategroups"
    _LIST_CLASS = TemplateGroupList

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def create(
        self, template_groups: Union[TemplateGroup, List[TemplateGroup]]
    ) -> Union[TemplateGroup, TemplateGroupList]:
        """`Create one or more template groups.`

        Args:
            template_groups (Union[TemplateGroup, List[TemplateGroup]])

        Returns:
            Union[TemplateGroup, TemplateGroupList]: Created template group(s)

        Examples:
            Create a new template group:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import TemplateGroup
                >>> c = CogniteClient()
                >>> template_group_1 = TemplateGroup("sdk-test-group", "This is a test group")
                >>> template_group_2 = TemplateGroup("sdk-test-group-2", "This is another test group")
                >>> c.templates.groups.create([template_group_1, template_group_2])
        """
        return self._create_multiple(items=template_groups)

    def upsert(
        self, template_groups: Union[TemplateGroup, List[TemplateGroup]]
    ) -> Union[TemplateGroup, TemplateGroupList]:
        """`Upsert one or more template groups.`
        Will overwrite existing template group(s) with the same external id(s).

        Args:
            template_groups (Union[TemplateGroup, List[TemplateGroup]])

        Returns:
            Union[TemplateGroup, TemplateGroupList]: Created template group(s)

        Examples:
            Upsert a template group:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import TemplateGroup
                >>> c = CogniteClient()
                >>> template_group_1 = TemplateGroup("sdk-test-group", "This is a test group")
                >>> template_group_2 = TemplateGroup("sdk-test-group-2", "This is another test group")
                >>> c.templates.groups.upsert([template_group_1, template_group_2])
        """
        path = self._RESOURCE_PATH + "/upsert"
        is_single = not isinstance(template_groups, list)
        if is_single:
            template_groups = [template_groups]
        updated = self._post(path, {"items": [item.dump(camel_case=True) for item in template_groups]}).json()["items"]
        res = self._LIST_CLASS._load(updated, cognite_client=self._cognite_client)
        if is_single:
            return res[0]
        return res

    def retrieve_multiple(self, external_ids: List[str], ignore_unknown_ids: bool = False) -> TemplateGroupList:
        """`Retrieve multiple template groups by external id.`

        Args:
            external_ids (List[str]): External IDs
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            TemplateGroupList: The requested template groups.

        Examples:
            Get template groups by external id:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.templates.groups.retrieve_multiple(external_ids=["abc", "def"])
        """
        return self._retrieve_multiple(external_ids=external_ids, ignore_unknown_ids=ignore_unknown_ids, wrap_ids=True)

    def list(self, limit: int = 25, owners: List[str] = None) -> TemplateGroupList:
        """`Lists template groups stored in the project based on a query filter given in the payload of this request.`
        Up to 1000 template groups can be retrieved in one operation.

        Args:
            owners (List[str]): Include template groups that have any of these values in their `owner` field.
            limit (int): Maximum number of template groups to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            TemplateGroupList: List of requested template groups

        Examples:
            List template groups:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> template_group_list = c.templates.groups.list(limit=5)
        """
        filter = {}
        if owners is not None:
            filter["owners"] = owners
        return self._list(method="POST", limit=limit, filter=filter, partitions=None, sort=None)

    def delete(
        self, external_ids: Union[str, List[str]], ignore_unknown_ids: bool = False
    ) -> Union[TemplateGroup, TemplateGroupList]:
        """`Delete one or more template groups.`

        Args:
            external_ids (Union[str, List[str]]): External ID or list of external ids
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            None

        Examples:
            Delete template groups by external id:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.groups.delete(external_ids=["a", "b"])
        """
        return self._delete_multiple(
            True, external_ids=external_ids, extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids}
        )


class TemplateGroupVersionsAPI(APIClient):
    _RESOURCE_PATH = "/templategroups/{}/versions"
    _LIST_CLASS = TemplateGroupVersionList

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def upsert(self, external_id: str, version: TemplateGroupVersion):
        """`Upsert a template group version.`
        A Template Group update supports specifying different conflict modes, which is used when an existing schema already exists.

        Patch -> It diffs the new schema with the old schema and fails if there are breaking changes.
        Update -> It sets the new schema as schema of a new version.
        Force -> It ignores breaking changes and replaces the old schema with the new schema.
        The default mode is "patch".

        Args:
            external_id (str): The external id of the template group.
            version (TemplateGroupVersion): The GraphQL schema of this version.

        Returns:
            TemplateGroupVersion: Created template group version

        Examples:
            create a new template group version modeling Covid-19:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import TemplateGroup
                >>> c = CogniteClient()
                >>> template_group = TemplateGroup("sdk-test-group", "This template group models Covid-19 spread")
                >>> c.templates.groups.create(template_group)
                >>> schema = '''
                >>>     type Demographics @template {
                >>>         "The amount of people"
                >>>         populationSize: Int,
                >>>         "The population growth rate"
                >>>         growthRate: Float,
                >>>     }
                >>>     type Country @template {
                >>>         name: String,
                >>>         demographics: Demographics,
                >>>         deaths: TimeSeries,
                >>>         confirmed: TimeSeries,
                >>>     }'''
                >>> template_group_version = TemplateGroupVersion(schema)
                >>> c.templates.versions.upsert(template_group.external_id, template_group_version)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id) + "/upsert"
        version = self._post(resource_path, version.dump(camel_case=True)).json()
        return TemplateGroupVersion._load(version)

    def list(
        self, external_id: str, limit: int = 25, min_version: Optional[int] = None, max_version: Optional[int] = None
    ) -> TemplateGroupVersionList:
        """`Lists versions of a specified template group.`
        Up to 1000 template group version can be retrieved in one operation.

        Args:
            external_id (str): The external id of the template group.
            limit (int): Maximum number of template group versions to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            min_version: (Optional[int]): Exclude versions with a version number smaller than this.
            max_version: (Optional[int]): Exclude versions with a version number larger than this.

        Returns:
            TemplateGroupVersionList: List of requested template group versions

        Examples:
            List template group versions:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> template_group_list = c.templates.versions.list("template-group-ext-id", limit=5)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id)
        filter = {}
        if min_version is not None:
            filter["minVersion"] = min_version
        if max_version is not None:
            filter["maxVersion"] = max_version
        return self._list(resource_path=resource_path, method="POST", limit=limit, filter=filter)

    def delete(self, external_id: str, version: int) -> None:
        """`Delete a template group version.`

        Args:
            external_id (Union[str, List[str]]): External ID of the template group.
            version (int): The version of the template group to delete.

        Returns:
            None

        Examples:
            Delete template groups by external id:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.versions.delete("sdk-test-group", 1)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id)
        self._post(resource_path + "/delete", {"version": version})


class TemplateInstancesAPI(APIClient):
    _RESOURCE_PATH = "/templategroups/{}/versions/{}/instances"
    _LIST_CLASS = TemplateInstanceList

    def create(
        self, external_id: str, version: int, instances: Union[TemplateInstance, List[TemplateInstance]]
    ) -> Union[TemplateInstance, TemplateInstanceList]:
        """`Create one or more template instances.`

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group to create instances for.
            instances (Union[TemplateInstance, List[TemplateInstance]]): The instances to create.

        Returns:
            Union[TemplateInstance, TemplateInstanceList]: Created template instance(s).

        Examples:
            create new template instances for Covid-19 spread:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import TemplateInstance
                >>> c = CogniteClient()
                >>> template_instance_1 = TemplateInstance(external_id="norway",
                >>>                               template_name="Country",
                >>>                               field_resolvers={
                >>>                                   "name": ConstantResolver("Norway"),
                >>>                                   "demographics": ConstantResolver("norway_demographics"),
                >>>                                   "deaths": ConstantResolver("Norway_deaths"),
                >>>                                   "confirmed": ConstantResolver("Norway_confirmed"),
                >>>                                   }
                >>>                               )
                >>> template_instance_2 = TemplateInstance(external_id="norway_demographics",
                >>>                               template_name="Demographics",
                >>>                               field_resolvers={
                >>>                                   "populationSize": ConstantResolver(5328000),
                >>>                                   "growthRate": ConstantResolver(value=0.02)
                >>>                                   }
                >>>                               )
                >>> c.templates.instances.create("sdk-test-group", 1, [template_instance_1, template_instance_2])
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._create_multiple(resource_path=resource_path, items=instances)

    def upsert(
        self, external_id: str, version: int, instances: Union[TemplateGroup, List[TemplateGroup]]
    ) -> Union[TemplateInstance, TemplateInstanceList]:
        """`Upsert one or more template instances.`
        Will overwrite existing instances.

        Args:
         external_id (str): The external id of the template group.
         version (int): The version of the template group to create instances for.
         instances (Union[TemplateInstance, List[TemplateInstance]]): The instances to create.

        Returns:
         Union[TemplateInstance, TemplateInstanceList]: Created template instance(s).

        Examples:
         create new template instances for Covid-19 spread:

             >>> from cognite.client import CogniteClient
             >>> from cognite.client.data_classes import TemplateInstance
             >>> c = CogniteClient()
             >>> template_instance_1 = TemplateInstance(external_id="norway",
             >>>        template_name="Country",
             >>>        field_resolvers={
             >>>            "name": ConstantResolver("Norway"),
             >>>            "demographics": ConstantResolver("norway_demographics"),
             >>>            "deaths": ConstantResolver("Norway_deaths"),
             >>>            "confirmed": ConstantResolver("Norway_confirmed"),
             >>>        }
             >>>    )
             >>> template_instance_2 = TemplateInstance(external_id="norway_demographics",
             >>>       template_name="Demographics",
             >>>       field_resolvers={
             >>>           "populationSize": ConstantResolver(5328000),
             >>>           "growthRate": ConstantResolver(0.02)
             >>>           }
             >>>   )
             >>> c.templates.instances.upsert("sdk-test-group", 1, [template_instance_1, template_instance_2])
        """
        if isinstance(instances, TemplateInstance):
            instances = [instances]
        resource_path = (
            utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version) + "/upsert"
        )
        updated = self._post(
            resource_path, {"items": [instance.dump(camel_case=True) for instance in instances]}
        ).json()["items"]
        res = TemplateInstanceList._load(updated, cognite_client=self._cognite_client)
        if len(res) == 1:
            return res[0]
        return res

    def update(
        self, external_id: str, version: int, item: Union[TemplateInstanceUpdate, List[TemplateInstanceUpdate]]
    ) -> Union[TemplateInstance, List[TemplateInstance]]:
        """`Update one or more template instances`
        Args:
            item (Union[TemplateInstanceUpdate, List[TemplateInstanceUpdate]]): Templates instance(s) to update

        Returns:
            Union[TemplateInstance, List[TemplateInstance]]: Updated template instance(s)

        Examples:
            Perform a partial update on a template instance::

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import TemplateInstanceUpdate
                >>> c = CogniteClient()
                >>> my_update = TemplateInstanceUpdate(external_id="test").field_resolvers.add({ "name": ConstantResolver("Norway") })
                >>> res = c.templates.instances.update("sdk-test-group", 1, my_update)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._update_multiple(items=item, resource_path=resource_path)

    def retrieve_multiple(
        self, external_id: str, version: int, external_ids: List[str], ignore_unknown_ids: bool = False
    ) -> TemplateInstanceList:
        """`Retrieve multiple template instances by external id.`

        Args:
            external_id (str): The template group to retrieve instances from.
            version (int): The version of the template group.
            external_ids (List[str]): External IDs of the instances.
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            TemplateInstanceList: The requested template groups.

        Examples:
            Get template instances by external id:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> res = c.templates.instances.retrieve_multiple(external_id="sdk-test-group", version=1, external_ids=["abc", "def"])
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._retrieve_multiple(
            resource_path=resource_path, external_ids=external_ids, ignore_unknown_ids=ignore_unknown_ids, wrap_ids=True
        )

    def list(
        self,
        external_id: str,
        version: int,
        limit: int = 25,
        data_set_ids: Optional[List[int]] = None,
        template_names: Optional[List[str]] = None,
    ) -> TemplateInstanceList:
        """`Lists instances in a template group.`
        Up to 1000 template instances can be retrieved in one operation.

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group.
            limit (int): Maximum number of template group versions to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.
            data_set_ids: (Optional[List[int]]): Only include instances which has one of these values in their `data_set_id` field.
            template_names: (Optional[List[str]]): Only include instances which has one of these values in their `template_name` field.

        Returns:
            TemplateInstanceList: List of requested template instances

        Examples:
            List template instances:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> template_instances_list = c.templates.instances.list("template-group-ext-id", 1, limit=5)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        filter = {}
        if data_set_ids is not None:
            filter["dataSetIds"] = data_set_ids
        if template_names is not None:
            filter["templateNames"] = template_names
        return self._list(resource_path=resource_path, method="POST", limit=limit, filter=filter)

    def delete(self, external_id: str, version: int, external_ids: List[str], ignore_unknown_ids: bool = False) -> None:
        """`Delete one or more template instances.`

        Args:
            external_id (Union[str, List[str]]): External ID of the template group.
            version (int): The version of the template group.
            external_ids (List[str]): The external ids of the template instances to delete
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            None

        Examples:
            Delete template groups by external id:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.instances.delete("sdk-test-group", 1, external_ids=["a", "b"])
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._delete_multiple(
            resource_path=resource_path,
            external_ids=external_ids,
            wrap_ids=True,
            extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids},
        )


class TemplateViewsAPI(APIClient):
    _RESOURCE_PATH = "/templategroups/{}/versions/{}/views"
    _LIST_CLASS = ViewList

    def create(self, external_id: str, version: int, views: Union[View, List[View]]) -> Union[View, ViewList]:
        """`Create one or more template views.`

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group to create views for.
            views (Union[View, List[View]]): The views to create.

        Returns:
            Union[View, ViewList]: Created view(s).

        Examples:
            Create new views:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import View
                >>> c = CogniteClient()
                >>> view = View(external_id="view",
                >>>             source=Source(
                >>>                 type: 'events',
                >>>                 filter: {
                >>>                     startTime: {
                >>>                         min: "$startTime"
                >>>                     },
                >>>                     type: "Test",
                >>>                 }
                >>>                 mappings: {
                >>>                     author: "metadata/author"
                >>>                 }
                >>>             )
                >>>        )
                >>> c.templates.views.create("sdk-test-group", 1, [view])
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._create_multiple(resource_path=resource_path, items=views)

    def upsert(self, external_id: str, version: int, views: Union[View, List[View]]) -> Union[View, ViewList]:
        """`Upsert one or more template views.`

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group to create views for.
            views (Union[View, List[View]]): The views to create.

        Returns:
            Union[View, ViewList]: Created view(s).

        Examples:
            Upsert new views:

                >>> from cognite.client import CogniteClient
                >>> from cognite.client.data_classes import View
                >>> c = CogniteClient()
                >>> view = View(external_id="view",
                >>>             source=Source(
                >>>                 type: 'events',
                >>>                 filter: {
                >>>                     startTime: {
                >>>                         min: "$startTime"
                >>>                     },
                >>>                     type: "Test",
                >>>                 }
                >>>                 mappings: {
                >>>                     author: "metadata/author"
                >>>                 }
                >>>             )
                >>>        )
                >>> c.templates.views.upsert("sdk-test-group", 1, [view])
        """
        if isinstance(views, View):
            views = [views]
        resource_path = (
            utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version) + "/upsert"
        )
        updated = self._post(resource_path, {"items": [view.dump(camel_case=True) for view in views]}).json()["items"]
        res = ViewList._load(updated, cognite_client=self._cognite_client)
        if len(res) == 1:
            return res[0]
        return res

    def resolve(
        self, external_id: str, version: int, view_external_id: str, input: Optional[Dict[str, any]], limit: int = 25
    ) -> ViewResolveList:
        """`Resolves a View.`
        It resolves the source specified in a View with the provided input and applies the mapping rules to the response.

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group.
            input (Optional[Dict[str, any]]): The input for the View.
            limit (int): Maximum number of views to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            ViewResolveList: The resolved items.

        Examples:
            Resolve view:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.views.resolve("template-group-ext-id", 1, "view", { "startTime": 10 }, limit=5)
        """
        url_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version) + "/resolve"
        return self._list(
            url_path=url_path,
            method="POST",
            cls=ViewResolveList,
            limit=limit,
            other_params={"externalId": view_external_id, "input": input},
        )

    def list(self, external_id: str, version: int, limit: int = 25) -> ViewList:
        """`Lists view in a template group.`
        Up to 1000 views can be retrieved in one operation.

        Args:
            external_id (str): The external id of the template group.
            version (int): The version of the template group.
            limit (int): Maximum number of views to return. Defaults to 25. Set to -1, float("inf") or None
                to return all items.

        Returns:
            ViewList: List of requested views

        Examples:
            List views:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.views.list("template-group-ext-id", 1, limit=5)
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._list(resource_path=resource_path, method="POST", limit=limit)

    def delete(
        self, external_id: str, version: int, view_external_id: Union[List[str], str], ignore_unknown_ids: bool = False
    ) -> None:
        """`Delete one or more views.`

        Args:
            external_id (Union[str, List[str]]): External ID of the template group.
            version (int): The version of the template group.
            view_external_id (Union[List[str], str]): The external ids of the views to delete
            ignore_unknown_ids (bool): Ignore external IDs that are not found rather than throw an exception.

        Returns:
            None

        Examples:
            Delete views by external id:

                >>> from cognite.client import CogniteClient
                >>> c = CogniteClient()
                >>> c.templates.views.delete("sdk-test-group", 1, external_id=["a", "b"])
        """
        resource_path = utils._auxiliary.interpolate_and_url_encode(self._RESOURCE_PATH, external_id, version)
        return self._delete_multiple(
            resource_path=resource_path,
            external_ids=view_external_id,
            wrap_ids=True,
            extra_body_fields={"ignoreUnknownIds": ignore_unknown_ids},
        )



import pytest

from cognite.client.exceptions import CogniteAPIError


@pytest.fixture
def cognite_client_with_wrong_base_url(cognite_client):
    url = cognite_client.config.base_url
    cognite_client.config.base_url = "https://cognitedata.com"
    yield cognite_client
    cognite_client.config.base_url = url


class TestCogniteClient:
    def test_wrong_base_url_resulting_in_301(self, cognite_client_with_wrong_base_url):
        with pytest.raises(CogniteAPIError):
            cognite_client_with_wrong_base_url.assets.list(limit=1)

    def test_get(self, cognite_client):
        res = cognite_client.get("/login/status")
        assert res.status_code == 200

    def test_post(self, cognite_client):
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.post("/login", json={})
        assert e.value.code == 404

    def test_put(self, cognite_client):
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.put("/login")
        assert e.value.code == 404

    def test_delete(self, cognite_client):
        with pytest.raises(CogniteAPIError) as e:
            cognite_client.delete("/login")
        assert e.value.code == 404



import re

import pytest

from cognite.client._api.iam import APIKeyList, GroupList, SecurityCategoryList, ServiceAccountList
from cognite.client.data_classes import APIKey, Group, SecurityCategory, ServiceAccount
from cognite.client.data_classes.iam import ProjectSpec, TokenInspection
from tests.utils import jsgz_load


@pytest.fixture
def mock_service_accounts(rsps, cognite_client):
    response_body = {
        "items": [{"name": "service@bla.com", "groups": [1, 2, 3], "id": 0, "isDeleted": False, "deletedTime": 0}]
    }
    url_pattern = re.compile(re.escape(cognite_client.iam._get_base_url_with_base_path()) + "/serviceaccounts.*")
    rsps.assert_all_requests_are_fired = False
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestServiceAccounts:
    def test_list(self, cognite_client, mock_service_accounts):
        res = cognite_client.iam.service_accounts.list()
        assert isinstance(res, ServiceAccountList)
        assert mock_service_accounts.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_create(self, cognite_client, mock_service_accounts):
        res = cognite_client.iam.service_accounts.create(ServiceAccount(name="service@bla.com", groups=[1, 2, 3]))
        assert isinstance(res, ServiceAccount)
        assert {"items": [{"name": "service@bla.com", "groups": [1, 2, 3]}]} == jsgz_load(
            mock_service_accounts.calls[0].request.body
        )
        assert mock_service_accounts.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_service_accounts):
        res = cognite_client.iam.service_accounts.create([ServiceAccount(name="service@bla.com", groups=[1, 2, 3])])
        assert isinstance(res, ServiceAccountList)
        assert {"items": [{"name": "service@bla.com", "groups": [1, 2, 3]}]} == jsgz_load(
            mock_service_accounts.calls[0].request.body
        )
        assert mock_service_accounts.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_service_accounts):
        res = cognite_client.iam.service_accounts.delete(1)
        assert {"items": [1]} == jsgz_load(mock_service_accounts.calls[0].request.body)
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_service_accounts):
        res = cognite_client.iam.service_accounts.delete([1])
        assert {"items": [1]} == jsgz_load(mock_service_accounts.calls[0].request.body)
        assert res is None


@pytest.fixture
def mock_api_keys(rsps, cognite_client):
    response_body = {"items": [{"id": 1, "serviceAccountId": 1, "createdTime": 0, "status": "ACTIVE"}]}
    url_pattern = re.compile(re.escape(cognite_client.iam._get_base_url_with_base_path()) + "/apikeys.*")
    rsps.assert_all_requests_are_fired = False
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestAPIKeys:
    def test_list(self, cognite_client, mock_api_keys):
        res = cognite_client.iam.api_keys.list()
        assert isinstance(res, APIKeyList)
        assert mock_api_keys.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_create(self, cognite_client, mock_api_keys):
        res = cognite_client.iam.api_keys.create(1)
        assert isinstance(res, APIKey)
        assert {"items": [{"serviceAccountId": 1}]} == jsgz_load(mock_api_keys.calls[0].request.body)
        assert mock_api_keys.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_api_keys):
        res = cognite_client.iam.api_keys.create([1])
        assert isinstance(res, APIKeyList)
        assert {"items": [{"serviceAccountId": 1}]} == jsgz_load(mock_api_keys.calls[0].request.body)
        assert mock_api_keys.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_api_keys):
        res = cognite_client.iam.api_keys.delete(1)
        assert {"items": [1]} == jsgz_load(mock_api_keys.calls[0].request.body)
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_api_keys):
        res = cognite_client.iam.api_keys.delete([1])
        assert {"items": [1]} == jsgz_load(mock_api_keys.calls[0].request.body)
        assert res is None


@pytest.fixture
def mock_groups(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "name": "Production Engineers",
                "sourceId": "b7c9a5a4-99c2-4785-bed3-5e6ad9a78603",
                "capabilities": [{"groupsAcl": {"actions": ["LIST"], "scope": {"all": {}}}}],
                "id": 0,
                "isDeleted": False,
                "deletedTime": 0,
            }
        ]
    }
    url_pattern = re.compile(re.escape(cognite_client.iam._get_base_url_with_base_path()) + "/groups.*")
    rsps.assert_all_requests_are_fired = False
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_group_service_account_response(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "name": "some-internal-service@apple.com",
                "groups": [1, 2, 3],
                "id": 0,
                "isDeleted": False,
                "deletedTime": 0,
            }
        ]
    }
    url_pattern = re.compile(
        re.escape(cognite_client.iam._get_base_url_with_base_path()) + "/groups/1/serviceaccounts.*"
    )
    rsps.assert_all_requests_are_fired = False
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_empty_response(rsps, cognite_client):
    response_body = {}
    url_pattern = re.compile(
        re.escape(cognite_client.iam._get_base_url_with_base_path()) + "/groups/1/serviceaccounts.*"
    )
    rsps.assert_all_requests_are_fired = False
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestGroups:
    def test_list(self, cognite_client, mock_groups):
        res = cognite_client.iam.groups.list()
        assert isinstance(res, GroupList)
        assert mock_groups.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_create(self, cognite_client, mock_groups):
        my_capabilities = [{"groupsAcl": {"actions": ["LIST"], "scope": {"all": {}}}}]
        my_group = Group(name="My Group", capabilities=my_capabilities)
        res = cognite_client.iam.groups.create(my_group)
        assert isinstance(res, Group)
        assert {"items": [{"name": "My Group", "capabilities": my_capabilities}]} == jsgz_load(
            mock_groups.calls[0].request.body
        )
        assert mock_groups.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_groups):
        res = cognite_client.iam.groups.create([1])
        assert isinstance(res, GroupList)
        assert {"items": [1]} == jsgz_load(mock_groups.calls[0].request.body)
        assert mock_groups.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_groups):
        res = cognite_client.iam.groups.delete(1)
        assert {"items": [1]} == jsgz_load(mock_groups.calls[0].request.body)
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_groups):
        res = cognite_client.iam.groups.delete([1])
        assert {"items": [1]} == jsgz_load(mock_groups.calls[0].request.body)
        assert res is None

    def test_list_service_accounts(self, cognite_client, mock_group_service_account_response):
        res = cognite_client.iam.groups.list_service_accounts(1)
        assert isinstance(res, ServiceAccountList)
        assert mock_group_service_account_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_add_service_account(self, cognite_client, mock_empty_response):
        res = cognite_client.iam.groups.add_service_account(1, 1)
        assert res is None
        assert {"items": [1]} == jsgz_load(mock_empty_response.calls[0].request.body)

    def test_add_service_account_multiple(self, cognite_client, mock_empty_response):
        res = cognite_client.iam.groups.add_service_account(1, [1])
        assert res is None
        assert {"items": [1]} == jsgz_load(mock_empty_response.calls[0].request.body)

    def test_remove_service_account(self, cognite_client, mock_empty_response):
        res = cognite_client.iam.groups.remove_service_account(1, 1)
        assert res is None
        assert {"items": [1]} == jsgz_load(mock_empty_response.calls[0].request.body)

    def test_remove_service_account_multiple(self, cognite_client, mock_empty_response):
        res = cognite_client.iam.groups.remove_service_account(1, [1])
        assert res is None
        assert {"items": [1]} == jsgz_load(mock_empty_response.calls[0].request.body)


@pytest.fixture
def mock_security_categories(rsps, cognite_client):
    response_body = {"items": [{"name": "bla", "id": 1}]}
    url_pattern = re.compile(re.escape(cognite_client.iam._get_base_url_with_base_path()) + "/securitycategories.*")
    rsps.assert_all_requests_are_fired = False
    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestSecurityCategories:
    def test_list(self, cognite_client, mock_security_categories):
        res = cognite_client.iam.security_categories.list()
        assert isinstance(res, SecurityCategoryList)
        assert mock_security_categories.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_create(self, cognite_client, mock_security_categories):
        res = cognite_client.iam.security_categories.create(SecurityCategory(name="My Category"))
        assert isinstance(res, SecurityCategory)
        assert {"items": [{"name": "My Category"}]} == jsgz_load(mock_security_categories.calls[0].request.body)
        assert mock_security_categories.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_security_categories):
        res = cognite_client.iam.security_categories.create([1])
        assert isinstance(res, SecurityCategoryList)
        assert {"items": [1]} == jsgz_load(mock_security_categories.calls[0].request.body)
        assert mock_security_categories.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_delete(self, cognite_client, mock_security_categories):
        res = cognite_client.iam.security_categories.delete(1)
        assert {"items": [1]} == jsgz_load(mock_security_categories.calls[0].request.body)
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_security_categories):
        res = cognite_client.iam.security_categories.delete([1])
        assert {"items": [1]} == jsgz_load(mock_security_categories.calls[0].request.body)
        assert res is None


@pytest.fixture
def mock_token_inspect(rsps, cognite_client):
    response_body = {
        "subject": "someSubject",
        "projects": [{"projectUrlName": "veryGoodUrlName", "groups": [1, 2, 3]}],
        "capabilities": [{"groupsAcl": {"actions": ["LIST"], "scope": {"all": {}}}}],
    }
    url_pattern = re.compile(
        re.escape(cognite_client.iam.token._get_base_url_with_base_path()) + "/api/v1/token/inspect"
    )
    rsps.assert_all_requests_are_fired = False
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


class TestTokenAPI:
    def test_token_inspect(self, cognite_client, mock_token_inspect):
        res = cognite_client.iam.token.inspect()
        assert isinstance(res, TokenInspection)
        assert res.subject == "someSubject"
        assert res.projects == [ProjectSpec(url_name="veryGoodUrlName", groups=[1, 2, 3])]
        assert res.capabilities == [{"groupsAcl": {"actions": ["LIST"], "scope": {"all": {}}}}]

    def test_token_inspection_dump(self):
        capabilities = [{"groupsAcl": {"actions": ["LIST"], "scope": {"all": {}}}}]
        groups = [1, 2, 3]
        obj = TokenInspection("subject", [ProjectSpec("urlName", groups)], capabilities)

        assert obj.dump() == {
            "subject": "subject",
            "projects": [{"url_name": "urlName", "groups": groups}],
            "capabilities": capabilities,
        }
        assert obj.dump(camel_case=True) == {
            "subject": "subject",
            "projects": [{"urlName": "urlName", "groups": groups}],
            "capabilities": capabilities,
        }



import os
from unittest import mock

import pytest
import responses

from cognite.client import CogniteClient
from cognite.client._api.assets import AssetsAPI
from cognite.client._api.data_sets import DataSetsAPI
from cognite.client._api.datapoints import DatapointsAPI
from cognite.client._api.entity_matching import EntityMatchingAPI
from cognite.client._api.events import EventsAPI
from cognite.client._api.files import FilesAPI
from cognite.client._api.iam import IAMAPI, APIKeysAPI, GroupsAPI, SecurityCategoriesAPI, ServiceAccountsAPI
from cognite.client._api.login import LoginAPI
from cognite.client._api.raw import RawAPI, RawDatabasesAPI, RawRowsAPI, RawTablesAPI
from cognite.client._api.relationships import RelationshipsAPI
from cognite.client._api.sequences import SequencesAPI, SequencesDataAPI
from cognite.client._api.three_d import (
    ThreeDAPI,
    ThreeDAssetMappingAPI,
    ThreeDFilesAPI,
    ThreeDModelsAPI,
    ThreeDRevisionsAPI,
)
from cognite.client._api.time_series import TimeSeriesAPI
from tests.utils import BASE_URL


@pytest.fixture
def rsps_with_login_mock():
    with responses.RequestsMock() as rsps:
        rsps.add(rsps.GET, "https://pypi.python.org/simple/cognite-sdk/#history", status=200, body="")
        rsps.add(
            rsps.GET,
            BASE_URL + "/login/status",
            status=200,
            json={"data": {"project": "test", "loggedIn": True, "user": "bla", "projectId": "bla", "apiKeyId": 1}},
        )
        yield rsps


@pytest.fixture
def mock_cognite_client():
    with mock.patch("cognite.client.CogniteClient") as client_mock:
        cog_client_mock = mock.MagicMock(spec=CogniteClient)
        cog_client_mock.time_series = mock.MagicMock(spec=TimeSeriesAPI)
        cog_client_mock.datapoints = mock.MagicMock(spec=DatapointsAPI)
        cog_client_mock.assets = mock.MagicMock(spec=AssetsAPI)
        cog_client_mock.events = mock.MagicMock(spec=EventsAPI)
        cog_client_mock.data_sets = mock.MagicMock(spec=DataSetsAPI)
        cog_client_mock.files = mock.MagicMock(spec=FilesAPI)
        cog_client_mock.login = mock.MagicMock(spec=LoginAPI)
        cog_client_mock.three_d = mock.MagicMock(spec=ThreeDAPI)
        cog_client_mock.three_d.models = mock.MagicMock(spec=ThreeDModelsAPI)
        cog_client_mock.three_d.revisions = mock.MagicMock(spec=ThreeDRevisionsAPI)
        cog_client_mock.three_d.files = mock.MagicMock(spec=ThreeDFilesAPI)
        cog_client_mock.three_d.asset_mappings = mock.MagicMock(spec=ThreeDAssetMappingAPI)
        cog_client_mock.iam = mock.MagicMock(spec=IAMAPI)
        cog_client_mock.iam.service_accounts = mock.MagicMock(spec=ServiceAccountsAPI)
        cog_client_mock.iam.api_keys = mock.MagicMock(spec=APIKeysAPI)
        cog_client_mock.iam.groups = mock.MagicMock(spec=GroupsAPI)
        cog_client_mock.iam.security_categories = mock.MagicMock(spec=SecurityCategoriesAPI)
        cog_client_mock.sequences = mock.MagicMock(spec=SequencesAPI)
        cog_client_mock.sequences.data = mock.MagicMock(spec=SequencesDataAPI)
        cog_client_mock.relationships = mock.MagicMock(spec=RelationshipsAPI)
        raw_mock = mock.MagicMock(spec=RawAPI)
        raw_mock.databases = mock.MagicMock(spec=RawDatabasesAPI)
        raw_mock.tables = mock.MagicMock(spec=RawTablesAPI)
        raw_mock.rows = mock.MagicMock(spec=RawRowsAPI)
        cog_client_mock.raw = raw_mock
        client_mock.return_value = cog_client_mock
        yield


@pytest.fixture
def mock_cognite_beta_client(mock_cognite_client):
    with mock.patch("cognite.client.beta.CogniteClient") as client_mock:
        cog_client_mock = mock.MagicMock(spec=CogniteClient)
        cog_client_mock.entity_matching = mock.MagicMock(spec=EntityMatchingAPI)
        client_mock.return_value = cog_client_mock
        yield


@pytest.fixture
def rsps():
    with responses.RequestsMock() as rsps:
        yield rsps


@pytest.fixture
def disable_gzip():
    os.environ["COGNITE_DISABLE_GZIP"] = "1"
    yield
    del os.environ["COGNITE_DISABLE_GZIP"]


def pytest_addoption(parser):
    parser.addoption(
        "--test-deps-only-core", action="store_true", default=False, help="Test only core deps are installed"
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--test-deps-only-core"):
        return
    skip_core = pytest.mark.skip(reason="need --test-deps-only-core option to run")
    for item in items:
        if "coredeps" in item.keywords:
            item.add_marker(skip_core)



from unittest.mock import patch

import pytest

from cognite.client import CogniteClient
from cognite.client.data_classes import ContextualizationJob
from tests.utils import set_env_var


@pytest.fixture()
def job():
    with set_env_var("COGNITE_API_KEY", "BLA"):
        return ContextualizationJob(job_id=123, status="Queued", cognite_client=CogniteClient())


def mock_update_status_running(self):
    self.status = "Running"
    return self.status


def mock_update_status_completed(self):
    self.status = "Completed"
    return self.status


class TestContextualizationJob:
    @patch("cognite.client.data_classes.ContextualizationJob.update_status", new=mock_update_status_running)
    def test_wait_for_completion_running(self, job):
        job.wait_for_completion(timeout=1)
        assert job.status == "Running"

    @patch("cognite.client.data_classes.ContextualizationJob.update_status", new=mock_update_status_completed)
    def test_wait_for_completion_completed(self, job):
        job.wait_for_completion(timeout=1)
        assert job.status == "Completed"



import pytest

from cognite.client import utils
from cognite.client.data_classes import Asset, Datapoint
from tests.utils import jsgz_load


@pytest.fixture
def mock_ts_by_ids_response(rsps, cognite_client):
    res = {
        "items": [
            {
                "id": 1,
                "externalId": "1",
                "name": "stringname",
                "isString": True,
                "metadata": {"metadata-key": "metadata-value"},
                "unit": "string",
                "assetId": 1,
                "isStep": True,
                "description": "string",
                "securityCategories": [0],
                "createdTime": 0,
                "lastUpdatedTime": 0,
            }
        ]
    }
    rsps.add(
        rsps.POST, cognite_client.time_series._get_base_url_with_base_path() + "/timeseries/byids", status=200, json=res
    )
    yield rsps


@pytest.fixture
def mock_asset_by_ids_response(rsps, cognite_client):
    res = {"items": [{"id": 1, "externalId": "1", "name": "assetname"}]}
    rsps.add(
        rsps.POST, cognite_client.time_series._get_base_url_with_base_path() + "/assets/byids", status=200, json=res
    )
    yield rsps


@pytest.fixture
def mock_count_dps_in_ts(mock_ts_by_ids_response, cognite_client):
    mock_ts_by_ids_response.add(
        mock_ts_by_ids_response.POST,
        cognite_client.time_series._get_base_url_with_base_path() + "/timeseries/data/list",
        status=200,
        json={
            "items": [
                {
                    "id": 1,
                    "externalId": "1",
                    "isString": False,
                    "isStep": False,
                    "datapoints": [{"timestamp": 1, "count": 10}, {"timestamp": 2, "count": 5}],
                }
            ]
        },
    )
    yield mock_ts_by_ids_response


@pytest.fixture
def mock_get_latest_dp_in_ts(mock_ts_by_ids_response, cognite_client):
    mock_ts_by_ids_response.add(
        mock_ts_by_ids_response.POST,
        cognite_client.time_series._get_base_url_with_base_path() + "/timeseries/data/latest",
        status=200,
        json={
            "items": [
                {
                    "id": 1,
                    "externalId": "1",
                    "isString": False,
                    "isStep": False,
                    "datapoints": [{"timestamp": 1, "value": 10}],
                }
            ]
        },
    )
    yield mock_ts_by_ids_response


@pytest.fixture
def mock_get_first_dp_in_ts(mock_ts_by_ids_response, cognite_client):
    mock_ts_by_ids_response.add(
        mock_ts_by_ids_response.POST,
        cognite_client.time_series._get_base_url_with_base_path() + "/timeseries/data/list",
        status=200,
        json={
            "items": [
                {
                    "id": 1,
                    "externalId": "1",
                    "isString": False,
                    "isStep": False,
                    "datapoints": [{"timestamp": 1, "value": 10}],
                }
            ]
        },
    )
    yield mock_ts_by_ids_response


class TestTimeSeries:
    def test_get_count(self, cognite_client, mock_count_dps_in_ts):
        now = utils.timestamp_to_ms("now")
        assert 15 == cognite_client.time_series.retrieve(id=1).count()
        assert "count" == jsgz_load(mock_count_dps_in_ts.calls[1].request.body)["aggregates"][0]
        assert 0 == jsgz_load(mock_count_dps_in_ts.calls[1].request.body)["start"]
        assert now <= jsgz_load(mock_count_dps_in_ts.calls[1].request.body)["end"]

    def test_get_latest(self, cognite_client, mock_get_latest_dp_in_ts):
        res = cognite_client.time_series.retrieve(id=1).latest()
        assert isinstance(res, Datapoint)
        assert Datapoint(timestamp=1, value=10) == res

    def test_get_first_datapoint(self, cognite_client, mock_get_first_dp_in_ts):
        now = utils.timestamp_to_ms("now")
        res = cognite_client.time_series.retrieve(id=1).first()
        assert isinstance(res, Datapoint)
        assert Datapoint(timestamp=1, value=10) == res
        assert 0 == jsgz_load(mock_get_first_dp_in_ts.calls[1].request.body)["start"]
        assert now <= jsgz_load(mock_get_first_dp_in_ts.calls[1].request.body)["end"]
        assert 1 == jsgz_load(mock_get_first_dp_in_ts.calls[1].request.body)["limit"]

    def test_asset(self, cognite_client, mock_ts_by_ids_response, mock_asset_by_ids_response):
        now = utils.timestamp_to_ms("now")
        asset = cognite_client.time_series.retrieve(id=1).asset()
        assert isinstance(asset, Asset)
        assert "assetname" == asset.name



import pytest

from cognite.client import utils
from cognite.client.data_classes import Row
from cognite.client.utils._auxiliary import random_string


@pytest.fixture(scope="session")
def new_database_with_table(cognite_client):
    db_name = "db_" + utils._auxiliary.random_string(10)
    table_name = "table_" + utils._auxiliary.random_string(10)
    db = cognite_client.raw.databases.create(db_name)
    table = cognite_client.raw.tables.create(db_name, table_name)
    yield db, table
    cognite_client.raw.databases.delete(name=db.name, recursive=True)


class TestRawDatabasesAPI:
    def test_list_databases(self, cognite_client):
        dbs = cognite_client.raw.databases.list()
        assert len(dbs) > 0

    def test_create_and_delete_database(self, new_database_with_table):
        pass


class TestRawTablesAPI:
    def test_list_tables(self, cognite_client):
        tables = cognite_client.raw.tables.list(db_name="test__database1")
        assert len(tables) == 3

    def test_create_and_delete_table(self, cognite_client, new_database_with_table):
        db, _ = new_database_with_table
        table_name = "table_" + utils._auxiliary.random_string(10)
        table = cognite_client.raw.tables.create(db.name, table_name)
        assert table in cognite_client.raw.tables.list(db.name)
        cognite_client.raw.tables.delete(db.name, table.name)
        assert table not in cognite_client.raw.tables.list(db.name)


class TestRawRowsAPI:
    def test_list_rows(self, cognite_client):
        rows = cognite_client.raw.rows.list(db_name="test__database1", table_name="test__table_1", limit=10000)
        assert 2000 == len(rows)
        assert 10 == len(rows[0].columns.keys())

    def test_list_rows_w_parallel_cursors(self, cognite_client):
        randstr = random_string(32)
        num_rows = 30000
        rows_to_insert = [Row(key=str(i), columns={"a": 1}) for i in range(num_rows)]
        cognite_client.raw.rows.insert(randstr, randstr, row=rows_to_insert, ensure_parent=True)

        rows = cognite_client.raw.rows.list(db_name=randstr, table_name=randstr, limit=num_rows)
        rows_par = cognite_client.raw.rows.list(db_name=randstr, table_name=randstr, limit=-1)

        assert num_rows == len(rows) == len(rows_par)
        assert 1 == len(rows[0].columns.keys()) == len(rows_par[0].columns.keys())

        assert {row.key for row in rows} == {row.key for row in rows_par}

    def test_list_rows_cols(self, cognite_client):
        rows_list = cognite_client.raw.rows.list(
            db_name="test__database1", table_name="test__table_1", limit=10, columns=["c1", "c2"]
        )
        assert 10 == len(rows_list)
        for row in rows_list:
            assert {"c1", "c2"} == set(row.columns.keys())

    def test_iter_rows_cols(self, cognite_client):
        rows = cognite_client.raw.rows(
            db_name="test__database1", table_name="test__table_1", limit=10, columns=["c1", "c2"]
        )
        assert 10 == len([x for x in rows])
        for row in rows:
            assert {"c1", "c2"} == set(row.columns.keys())

    def test_retrieve_row(self, cognite_client):
        row = cognite_client.raw.rows.retrieve(db_name="test__database1", table_name="test__table_1", key="1")
        assert {"c{}".format(i): "1_{}".format(i) for i in range(10)} == row.columns

    def test_insert_and_delete_rows(self, cognite_client, new_database_with_table):
        db, table = new_database_with_table
        rows = {"r1": {"c1": "v1", "c2": "v1"}, "r2": {"c1": "v2", "c2": "v2"}}
        cognite_client.raw.rows.insert(db.name, table.name, rows)
        assert 2 == len(table.rows())
        cognite_client.raw.rows.delete(db.name, table.name, ["r1", "r2"])
        assert 0 == len(table.rows())

    @pytest.mark.dsl
    def test_insert_and_retrieve_dataframe(self, cognite_client, new_database_with_table):
        import pandas as pd

        db, table = new_database_with_table
        data = {"a": {"r1": 1, "r2": 1, "r3": 1}, "b": {"r1": None, "r2": None, "r3": None}}

        df = pd.DataFrame.from_dict(data)
        cognite_client.raw.rows.insert_dataframe(db.name, table.name, df)
        retrieved_df = cognite_client.raw.rows.retrieve_dataframe(db.name, table.name)

        pd.testing.assert_frame_equal(df.sort_index(), retrieved_df.sort_index())
        assert retrieved_df.to_dict() == data



import random

import pytest

from cognite.client.data_classes import (
    ContextualizationJob,
    EntityMatchingModel,
    EntityMatchingModelList,
    EntityMatchingModelUpdate,
)
from cognite.client.data_classes.contextualization import ContextualizationJobList
from cognite.client.exceptions import CogniteAPIError


@pytest.fixture
def fitted_model(cognite_client):
    extid = "abc" + str(random.randint(1, 1000000000))
    try:
        cognite_client.entity_matching.delete(external_id=extid)
    except CogniteAPIError as e:
        if e.code != 404:
            raise
    entities_from = [{"id": 1, "name": "xx-yy"}]
    entities_to = [{"id": 2, "bloop": "yy"}, {"id": 3, "bloop": "zz"}]
    model = cognite_client.entity_matching.fit(
        sources=entities_from,
        targets=entities_to,
        true_matches=[{"sourceId": 1, "targetId": 2}],
        feature_type="bigram",
        match_fields=[("name", "bloop")],
        external_id=extid,
    )
    yield model
    cognite_client.entity_matching.delete(id=model.id)


class TestEntityMatchingIntegration:
    def test_fit_retrieve_update(self, cognite_client, fitted_model):
        assert isinstance(fitted_model, EntityMatchingModel)
        assert "Queued" == fitted_model.status

        job = fitted_model.predict(sources=[{"name": "foo-bar"}], targets=[{"bloop": "foo-42"}])
        assert "Completed" == fitted_model.status
        assert isinstance(job, ContextualizationJob)
        assert "Queued" == job.status
        assert {"matches", "source"} == set(job.result["items"][0].keys()) - {"matchFrom"}
        assert "Completed" == job.status

        job = fitted_model.predict()
        assert isinstance(job, ContextualizationJob)
        assert "Queued" == job.status
        assert {"matches", "source"} == set(job.result["items"][0].keys()) - {"matchFrom"}
        assert "Completed" == job.status

        # Retrieve model
        model = cognite_client.entity_matching.retrieve(id=fitted_model.id)
        assert model.classifier == "randomforest"
        assert model.feature_type == "bigram"
        assert model.match_fields == [{"source": "name", "target": "bloop"}]

        # Retrieve models
        models = cognite_client.entity_matching.retrieve_multiple(ids=[model.id, model.id])
        assert 2 == len(models)
        assert model == models[0]
        assert model == models[1]

        # Update model
        model.name = "new_name"
        updated_model = cognite_client.entity_matching.update(model)
        assert type(updated_model) == EntityMatchingModel
        assert updated_model.name == "new_name"

        updated_model2 = cognite_client.entity_matching.update(
            EntityMatchingModelUpdate(id=model.id).description.set("new description")
        )
        assert type(updated_model2) == EntityMatchingModel
        assert updated_model2.description == "new description"

    def test_refit(self, cognite_client, fitted_model):
        new_model = fitted_model.refit(true_matches=[(1, 3)])
        assert new_model.id is not None
        assert new_model.id != fitted_model.id
        assert "Completed" == fitted_model.status
        assert isinstance(new_model, EntityMatchingModel)
        assert "Queued" == new_model.status

        job = new_model.predict(sources=[{"name": "foo-bar"}], targets=[{"bloop": "foo-42"}])
        assert {"matches", "source"} == set(job.result["items"][0].keys()) - {"matchFrom"}
        assert "Completed" == job.status
        cognite_client.entity_matching.delete(id=new_model.id)

    def test_true_match_formats(self, cognite_client):
        entities_from = [{"id": 1, "name": "xx-yy"}]
        entities_to = [{"id": 2, "name": "yy"}, {"id": 3, "externalId": "aa", "name": "xx"}]
        model = cognite_client.entity_matching.fit(
            sources=entities_from, targets=entities_to, true_matches=[{"sourceId": 1, "targetExternalId": "aa"}, (1, 2)]
        )
        assert isinstance(model, EntityMatchingModel)
        assert "Queued" == model.status
        cognite_client.entity_matching.delete(id=model.id)

    def test_extra_options(self, cognite_client):
        entities_from = [{"id": 1, "name": "xx-yy"}]
        entities_to = [{"id": 2, "name": "yy"}, {"id": 3, "name": "xx", "missing": "yy"}]
        model = cognite_client.entity_matching.fit(
            sources=entities_from,
            targets=entities_to,
            true_matches=[(1, 2)],
            feature_type="bigram",
            match_fields=[("name", "missing")],
            ignore_missing_fields=True,
            classifier="LogisticRegression",
            name="my_bigram_logReg_model",
            description="My model with bigram features",
        )
        assert isinstance(model, EntityMatchingModel)
        assert "Queued" == model.status
        job = model.predict()
        assert {"matches", "source"} == set(job.result["items"][0].keys()) - {"matchFrom"}

        cognite_client.entity_matching.delete(id=model.id)

    def test_list(self, cognite_client):
        models_list = cognite_client.entity_matching.list()
        assert len(models_list) > 0
        assert type(models_list) == EntityMatchingModelList
        assert all([type(x) == EntityMatchingModel for x in models_list])
        # Add filter
        models_list = cognite_client.entity_matching.list(feature_type="bigram")
        assert set([model.feature_type for model in models_list]) == {"bigram"}

    def test_list_jobs(self, cognite_client):
        jobs_list = cognite_client.entity_matching.list_jobs()
        assert len(jobs_list) > 0
        assert type(jobs_list) == ContextualizationJobList
        assert all([type(x) == ContextualizationJob for x in jobs_list])

    def test_direct_predict(self, cognite_client, fitted_model):
        job = cognite_client.entity_matching.predict(external_id=fitted_model.external_id)
        job2 = cognite_client.entity_matching.predict(id=fitted_model.id)
        assert isinstance(job, ContextualizationJob)
        assert "Queued" == job.status
        assert isinstance(job2, ContextualizationJob)

    def test_direct_refit(self, cognite_client, fitted_model):
        new_model = cognite_client.entity_matching.refit(external_id=fitted_model.external_id, true_matches=[(1, 3)])
        new_model2 = cognite_client.entity_matching.refit(id=fitted_model.id, true_matches=[(1, 3)])
        assert isinstance(new_model, EntityMatchingModel)
        assert isinstance(new_model2, EntityMatchingModel)
        cognite_client.entity_matching.delete(id=[new_model.id, new_model2.id])



import re
from datetime import datetime, timedelta
from unittest import mock

import numpy
import pandas
import pytest

from cognite.client import utils
from cognite.client.data_classes import DatapointsList, DatapointsQuery, TimeSeries
from cognite.client.exceptions import CogniteAPIError
from cognite.client.utils._time import timestamp_to_ms
from tests.utils import set_request_limit


@pytest.fixture(scope="session")
def test_time_series(cognite_client):
    eids = ["test__constant_%d_with_noise" % i for i in range(10)]
    ts = cognite_client.time_series.retrieve_multiple(external_ids=eids, ignore_unknown_ids=True)
    yield {int(re.match(r"test__constant_(\d+)_with_noise", t.name).group(1)): t for t in ts}


@pytest.fixture(scope="session")
def new_ts(cognite_client):
    ts = cognite_client.time_series.create(TimeSeries())
    yield ts
    cognite_client.time_series.delete(id=ts.id)
    assert cognite_client.time_series.retrieve(ts.id) is None


def has_duplicates(df: pandas.DataFrame):
    return df.duplicated().any()


def has_correct_timestamp_spacing(df: pandas.DataFrame, granularity: str):
    timestamps = df.index.values.astype("datetime64[ms]").astype("int64")
    deltas = numpy.diff(timestamps, 1)
    granularity_ms = utils._time.granularity_to_ms(granularity)
    return (deltas != 0).all() and (deltas % granularity_ms == 0).all()


@pytest.fixture
def post_spy(cognite_client):
    with mock.patch.object(cognite_client.datapoints, "_post", wraps=cognite_client.datapoints._post) as _:
        yield


class TestDatapointsAPI:
    def test_retrieve(self, cognite_client, test_time_series):
        ts = test_time_series[0]
        dps = cognite_client.datapoints.retrieve(id=ts.id, start="1d-ago", end="now")
        assert len(dps) > 0

    def test_retrieve_unknown(self, cognite_client, test_time_series):
        ts = test_time_series[0]
        dps = cognite_client.datapoints.retrieve(id=[ts.id] + [42], start="1d-ago", end="now", ignore_unknown_ids=True)
        assert 1 == len(dps)

    def test_retrieve_all_unknown(self, cognite_client, test_time_series):
        ts = test_time_series[0]
        dps = cognite_client.datapoints.retrieve(
            id=[42], external_id="missing", start="1d-ago", end="now", ignore_unknown_ids=True
        )
        assert isinstance(dps, DatapointsList)
        assert 0 == len(dps)

    def test_retrieve_all_unknown_single(self, cognite_client, test_time_series):
        ts = test_time_series[0]
        dps = cognite_client.datapoints.retrieve(
            external_id="missing", start="1d-ago", end="now", ignore_unknown_ids=True
        )
        assert dps is None

    def test_retrieve_multiple(self, cognite_client, test_time_series):
        ids = [test_time_series[0].id, test_time_series[1].id, {"id": test_time_series[2].id, "aggregates": ["max"]}]

        dps = cognite_client.datapoints.retrieve(
            id=ids, start="6h-ago", end="now", aggregates=["min"], granularity="1s"
        )
        df = dps.to_pandas(column_names="id")
        assert "{}|min".format(test_time_series[0].id) in df.columns
        assert "{}|min".format(test_time_series[1].id) in df.columns
        assert "{}|max".format(test_time_series[2].id) in df.columns
        assert 0 < df.shape[0]
        assert 3 == df.shape[1]
        assert has_correct_timestamp_spacing(df, "1s")
        for dpl in dps:
            assert dpl.is_step is not None
            assert dpl.is_string is not None

    def test_retrieve_nothing(self, cognite_client, test_time_series):
        dpl = cognite_client.datapoints.retrieve(id=test_time_series[0].id, start=0, end=1)
        assert 0 == len(dpl)
        assert dpl.is_step is not None
        assert dpl.is_string is not None

    def test_retrieve_multiple_with_exception(self, cognite_client, test_time_series):
        with pytest.raises(CogniteAPIError):
            cognite_client.datapoints.retrieve(
                id=[test_time_series[0].id, test_time_series[1].id, 0],
                start="1m-ago",
                end="now",
                aggregates=["min"],
                granularity="1s",
            )

    def test_stop_pagination_in_time(self, cognite_client, test_time_series, post_spy):
        lim = 152225
        ts = test_time_series[0]
        dps = cognite_client.datapoints.retrieve(id=ts.id, start=0, end="now", limit=lim)
        assert lim == len(dps)
        # first page 100k, counts 1, paginate 2 windows (+1 potential for 1st day uncertainty)
        assert 3 <= cognite_client.datapoints._post.call_count <= 4

    def test_retrieve_include_outside_points(self, cognite_client, test_time_series):
        ts = test_time_series[0]
        start = utils._time.timestamp_to_ms("6h-ago")
        end = utils._time.timestamp_to_ms("1h-ago")
        dps_wo_outside = cognite_client.datapoints.retrieve(
            id=ts.id, start=start, end=end, include_outside_points=False
        )
        dps_w_outside = cognite_client.datapoints.retrieve(id=ts.id, start=start, end=end, include_outside_points=True)
        assert not has_duplicates(dps_w_outside.to_pandas())
        assert len(dps_wo_outside) + 1 <= len(dps_w_outside) <= len(dps_wo_outside) + 2

    def test_retrieve_include_outside_points_paginate_no_outside(self, cognite_client, test_time_series, post_spy):
        ts = test_time_series[0]
        start = utils._time.timestamp_to_ms("156w-ago")
        end = utils._time.timestamp_to_ms("1h-ago")
        test_lim = 250
        dps_non_outside = cognite_client.datapoints.retrieve(id=ts.id, start=start, end=end, limit=1234)

        tmp = cognite_client.datapoints._DPS_LIMIT
        cognite_client.datapoints._DPS_LIMIT = test_lim
        dps = cognite_client.datapoints.retrieve(
            id=ts.id, start=start, end=end, include_outside_points=True, limit=1234
        )
        cognite_client.datapoints._DPS_LIMIT = tmp
        assert len(dps) == len(dps_non_outside)
        assert not has_duplicates(dps.to_pandas())
        ts_outside = set(dps.timestamp) - set(dps_non_outside.timestamp)
        assert 0 == len(ts_outside)
        assert set(dps.timestamp) == set(dps_non_outside.timestamp)

    def test_retrieve_include_outside_points_paginate_outside_exists(self, cognite_client, test_time_series, post_spy):
        ts = test_time_series[0]
        start = utils._time.timestamp_to_ms("12h-ago")
        end = utils._time.timestamp_to_ms("1h-ago")
        test_lim = 2500
        dps_non_outside = cognite_client.datapoints.retrieve(id=ts.id, start=start, end=end)
        tmp = cognite_client.datapoints._DPS_LIMIT
        cognite_client.datapoints._DPS_LIMIT = test_lim
        dps = cognite_client.datapoints.retrieve(id=ts.id, start=start, end=end, include_outside_points=True)
        cognite_client.datapoints._DPS_LIMIT = tmp
        ts_outside = set(dps.timestamp) - set(dps_non_outside.timestamp)
        assert 2 == len(ts_outside)
        for ts in ts_outside:
            assert ts < timestamp_to_ms(start) or ts >= timestamp_to_ms(end)
        assert set(dps.timestamp) - ts_outside == set(dps_non_outside.timestamp)

    def test_retrieve_dataframe(self, cognite_client, test_time_series):
        ts = test_time_series[0]
        df = cognite_client.datapoints.retrieve_dataframe(
            id=ts.id, start="6h-ago", end="now", aggregates=["average"], granularity="1s"
        )
        assert df.shape[0] > 0
        assert df.shape[1] == 1
        assert has_correct_timestamp_spacing(df, "1s")

    def test_retrieve_dataframe_missing(self, cognite_client, test_time_series):
        ts = test_time_series[0]
        df = cognite_client.datapoints.retrieve_dataframe(
            id=ts.id,
            external_id="missing",
            start="6h-ago",
            end="now",
            aggregates=["average"],
            granularity="1s",
            ignore_unknown_ids=True,
        )
        assert df.shape[0] > 0
        assert df.shape[1] == 1

    def test_retrieve_string(self, cognite_client):
        dps = cognite_client.datapoints.retrieve(external_id="test__string_b", start=1563000000000, end=1564100000000)
        assert len(dps) > 100000

    def test_query(self, cognite_client, test_time_series):
        dps_query1 = DatapointsQuery(id=test_time_series[0].id, start="6h-ago", end="now")
        dps_query2 = DatapointsQuery(id=test_time_series[1].id, start="3h-ago", end="now")
        dps_query3 = DatapointsQuery(
            id=test_time_series[2].id, start="1d-ago", end="now", aggregates=["average"], granularity="1h"
        )

        res = cognite_client.datapoints.query([dps_query1, dps_query2, dps_query3])
        assert len(res) == 3
        assert len(res[2][0]) < len(res[1][0]) < len(res[0][0])

    def test_query_unknown(self, cognite_client, test_time_series):
        dps_query1 = DatapointsQuery(id=test_time_series[0].id, start="6h-ago", end="now", ignore_unknown_ids=True)
        dps_query2 = DatapointsQuery(id=123, start="3h-ago", end="now", ignore_unknown_ids=True)
        dps_query3 = DatapointsQuery(
            external_id="missing time series",
            start="1d-ago",
            end="now",
            aggregates=["average"],
            granularity="1h",
            ignore_unknown_ids=True,
        )
        res = cognite_client.datapoints.query([dps_query1, dps_query2, dps_query3])
        assert len(res) == 3
        assert len(res[0]) == 1
        assert len(res[0][0]) > 0
        assert len(res[1]) == 0
        assert len(res[2]) == 0

    def test_retrieve_latest(self, cognite_client, test_time_series):
        ids = [test_time_series[0].id, test_time_series[1].id]
        res = cognite_client.datapoints.retrieve_latest(id=ids)
        for dps in res:
            assert 1 == len(dps)

    def test_retrieve_latest_unknown(self, cognite_client, test_time_series):
        ids = [test_time_series[0].id, test_time_series[1].id, 42, 1337]
        res = cognite_client.datapoints.retrieve_latest(id=ids, ignore_unknown_ids=True)
        assert 2 == len(res)
        for dps in res:
            assert 1 == len(dps)

    def test_retrieve_latest_many(self, cognite_client, test_time_series, post_spy):
        ids = [
            t.id for t in cognite_client.time_series.list(limit=12) if not t.security_categories
        ]  # more than one page
        assert len(ids) > 10
        tmp = cognite_client.datapoints._RETRIEVE_LATEST_LIMIT
        cognite_client.datapoints._RETRIEVE_LATEST_LIMIT = 10
        res = cognite_client.datapoints.retrieve_latest(id=ids)
        cognite_client.datapoints._RETRIEVE_LATEST_LIMIT = tmp
        assert len(ids) == len(res)
        assert 2 == cognite_client.datapoints._post.call_count
        for i, dps in enumerate(res):
            assert len(dps) <= 1  # could be empty
            assert ids[i] == res[i].id

    def test_retrieve_latest_before(self, cognite_client, test_time_series):
        ts = test_time_series[0]
        res = cognite_client.datapoints.retrieve_latest(id=ts.id, before="1h-ago")
        assert 1 == len(res)
        assert res[0].timestamp < utils._time.timestamp_to_ms("1h-ago")

    def test_insert(self, cognite_client, new_ts, post_spy):
        datapoints = [(datetime(year=2018, month=1, day=1, hour=1, minute=i), i) for i in range(60)]
        with set_request_limit(cognite_client.datapoints, 30):
            cognite_client.datapoints.insert(datapoints, id=new_ts.id)
        assert 2 == cognite_client.datapoints._post.call_count

    def test_insert_copy(self, cognite_client, test_time_series, new_ts, post_spy):
        data = cognite_client.datapoints.retrieve(id=test_time_series[0].id, start="600d-ago", end="now", limit=100)
        assert 100 == len(data)
        cognite_client.datapoints.insert(data, id=new_ts.id)
        assert 2 == cognite_client.datapoints._post.call_count

    def test_insert_pandas_dataframe(self, cognite_client, new_ts, post_spy):
        start = datetime(2018, 1, 1)
        x = pandas.DatetimeIndex([start + timedelta(days=d) for d in range(100)])
        y = numpy.random.normal(0, 1, 100)
        df = pandas.DataFrame({new_ts.id: y}, index=x)
        with set_request_limit(cognite_client.datapoints, 50):
            cognite_client.datapoints.insert_dataframe(df)
        assert 2 == cognite_client.datapoints._post.call_count

    def test_delete_range(self, cognite_client, new_ts):
        cognite_client.datapoints.delete_range(start="2d-ago", end="now", id=new_ts.id)

    def test_delete_ranges(self, cognite_client, new_ts):
        cognite_client.datapoints.delete_ranges([{"start": "2d-ago", "end": "now", "id": new_ts.id}])

    def test_retrieve_dataframe_dict(self, cognite_client, test_time_series):

        dfd = cognite_client.datapoints.retrieve_dataframe_dict(
            id=[test_time_series[0].id, 42],
            external_id=["missing time series", test_time_series[1].external_id],
            aggregates=["count", "interpolation"],
            start=0,
            end="now",
            limit=100,
            granularity="1m",
            ignore_unknown_ids=True,
        )
        assert isinstance(dfd, dict)
        assert 2 == len(dfd.keys())
        assert dfd["interpolation"].shape[0] > 0
        assert dfd["interpolation"].shape[1] == 2

        assert dfd["count"].shape[0] > 0
        assert dfd["count"].shape[1] == 2



import re

import pytest

from cognite.client._api.events import Event, EventList, EventUpdate
from cognite.client.data_classes import (
    AggregateResult,
    AggregateUniqueValuesResult,
    EndTimeFilter,
    EventFilter,
    TimestampRange,
)
from tests.utils import jsgz_load


@pytest.fixture
def mock_events_response(rsps, cognite_client):
    response_body = {
        "items": [
            {
                "externalId": "string",
                "startTime": 0,
                "endTime": 0,
                "description": "string",
                "metadata": {"metadata-key": "metadata-value"},
                "assetIds": [1],
                "source": "string",
                "id": 1,
                "lastUpdatedTime": 0,
            }
        ]
    }

    url_pattern = re.compile(re.escape(cognite_client.events._get_base_url_with_base_path()) + "/.+")
    rsps.assert_all_requests_are_fired = False

    rsps.add(rsps.POST, url_pattern, status=200, json=response_body)
    rsps.add(rsps.GET, url_pattern, status=200, json=response_body)
    yield rsps


@pytest.fixture
def mock_count_aggregate_response(rsps, cognite_client):
    url_pattern = re.compile(re.escape(cognite_client.events._get_base_url_with_base_path()) + "/events/aggregate")
    rsps.add(rsps.POST, url_pattern, status=200, json={"items": [{"count": 10}]})
    yield rsps


@pytest.fixture
def mock_aggregate_unique_values_response(rsps, cognite_client):
    url_pattern = re.compile(re.escape(cognite_client.events._get_base_url_with_base_path()) + "/events/aggregate")
    rsps.add(rsps.POST, url_pattern, status=200, json={"items": [{"count": 5, "value": "WORKORDER"}]})
    yield rsps


class TestEvents:
    def test_retrieve_single(self, cognite_client, mock_events_response):
        res = cognite_client.events.retrieve(id=1)
        assert isinstance(res, Event)
        assert mock_events_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_retrieve_multiple(self, cognite_client, mock_events_response):
        res = cognite_client.events.retrieve_multiple(ids=[1])
        assert isinstance(res, EventList)
        assert mock_events_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list(self, cognite_client, mock_events_response):
        res = cognite_client.events.list(source="bla")
        assert "bla" == jsgz_load(mock_events_response.calls[0].request.body)["filter"]["source"]
        assert mock_events_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list_partitions(self, cognite_client, mock_events_response):
        cognite_client.events.list(partitions=13, limit=float("inf"))
        assert 13 == len(mock_events_response.calls)

    def test_list_with_dataset_ids(self, cognite_client, mock_events_response):
        cognite_client.events.list(source="bla", data_set_ids=[1], data_set_external_ids=["x"])
        assert [{"id": 1}, {"externalId": "x"}] == jsgz_load(mock_events_response.calls[0].request.body)["filter"][
            "dataSetIds"
        ]

    def test_list_sorting(self, cognite_client, mock_events_response):
        res = cognite_client.events.list(sort=["startTime:desc"])
        assert ["startTime:desc"] == jsgz_load(mock_events_response.calls[0].request.body)["sort"]
        assert mock_events_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_list_sorting_combined_with_partitions(self, cognite_client, mock_events_response):
        with pytest.raises(ValueError):
            cognite_client.events.list(sort=["startTime:desc"], partitions=10)

    def test_list_with_time_dict(self, cognite_client, mock_events_response):
        cognite_client.events.list(start_time={"min": 20})
        assert 20 == jsgz_load(mock_events_response.calls[0].request.body)["filter"]["startTime"]["min"]
        assert "max" not in jsgz_load(mock_events_response.calls[0].request.body)["filter"]["startTime"]

    def test_list_with_timestamp_range(self, cognite_client, mock_events_response):
        cognite_client.events.list(start_time=TimestampRange(min=20))
        assert 20 == jsgz_load(mock_events_response.calls[0].request.body)["filter"]["startTime"]["min"]
        assert "max" not in jsgz_load(mock_events_response.calls[0].request.body)["filter"]["startTime"]

    def test_count_aggregate(self, cognite_client, mock_count_aggregate_response):
        res = cognite_client.events.aggregate(filter={"type": "WORKORDER"})
        assert isinstance(res[0], AggregateResult)
        assert res[0].count == 10

    def test_aggregate_unique_values(self, cognite_client, mock_aggregate_unique_values_response):
        res = cognite_client.events.aggregate_unique_values(filter={"type": "WORKORDER"}, fields=["subtype"])
        assert isinstance(res[0], AggregateUniqueValuesResult)
        assert res[0].count == 5
        assert res[0].value == "WORKORDER"

    def test_call_root(self, cognite_client, mock_events_response):
        list(
            cognite_client.events.__call__(
                root_asset_ids=[23], root_asset_external_ids=["a", "b"], asset_subtree_external_ids=["a"], limit=10
            )
        )
        calls = mock_events_response.calls
        assert 1 == len(calls)
        assert {
            "cursor": None,
            "limit": 10,
            "filter": {
                "rootAssetIds": [{"id": 23}, {"externalId": "a"}, {"externalId": "b"}],
                "assetSubtreeIds": [{"externalId": "a"}],
            },
        } == jsgz_load(calls[0].request.body)

    def test_list_root_ids_list(self, cognite_client, mock_events_response):
        cognite_client.events.list(root_asset_ids=[1, 2], limit=10)
        calls = mock_events_response.calls
        assert 1 == len(calls)
        assert {"cursor": None, "limit": 10, "filter": {"rootAssetIds": [{"id": 1}, {"id": 2}]}} == jsgz_load(
            calls[0].request.body
        )

    def test_list_root_extids_list(self, cognite_client, mock_events_response):
        cognite_client.events.list(root_asset_external_ids=["1", "2"], limit=10)
        calls = mock_events_response.calls
        assert 1 == len(calls)
        assert {
            "cursor": None,
            "limit": 10,
            "filter": {"rootAssetIds": [{"externalId": "1"}, {"externalId": "2"}]},
        } == jsgz_load(calls[0].request.body)

    def test_list_subtree(self, cognite_client, mock_events_response):
        res = cognite_client.events.list(limit=10, asset_subtree_external_ids=["a"], asset_subtree_ids=[1, 2])
        calls = mock_events_response.calls
        assert 1 == len(calls)
        assert {
            "cursor": None,
            "limit": 10,
            "filter": {"assetSubtreeIds": [{"id": 1}, {"id": 2}, {"externalId": "a"}]},
        } == jsgz_load(calls[0].request.body)

    def test_list_ongoing_wrong_signature(self, cognite_client):
        with pytest.raises(ValueError):
            cognite_client.events.list(end_time=EndTimeFilter(is_null=True, max=100))

    def test_create_single(self, cognite_client, mock_events_response):
        res = cognite_client.events.create(Event(external_id="1"))
        assert isinstance(res, Event)
        assert mock_events_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_create_multiple(self, cognite_client, mock_events_response):
        res = cognite_client.events.create([Event(external_id="1")])
        assert isinstance(res, EventList)
        assert mock_events_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_iter_single(self, cognite_client, mock_events_response):
        for event in cognite_client.events:
            assert mock_events_response.calls[0].response.json()["items"][0] == event.dump(camel_case=True)

    def test_iter_chunk(self, cognite_client, mock_events_response):
        for events in cognite_client.events(chunk_size=1):
            assert mock_events_response.calls[0].response.json()["items"] == events.dump(camel_case=True)

    def test_delete_single(self, cognite_client, mock_events_response):
        res = cognite_client.events.delete(id=1)
        assert {"ignoreUnknownIds": False, "items": [{"id": 1}]} == jsgz_load(
            mock_events_response.calls[0].request.body
        )
        assert res is None

    def test_delete_multiple(self, cognite_client, mock_events_response):
        res = cognite_client.events.delete(id=[1], ignore_unknown_ids=True)
        assert {"ignoreUnknownIds": True, "items": [{"id": 1}]} == jsgz_load(mock_events_response.calls[0].request.body)
        assert res is None

    def test_update_with_resource_class(self, cognite_client, mock_events_response):
        res = cognite_client.events.update(Event(id=1))
        assert isinstance(res, Event)
        assert mock_events_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_with_update_class(self, cognite_client, mock_events_response):
        res = cognite_client.events.update(EventUpdate(id=1).description.set("blabla"))
        assert isinstance(res, Event)
        assert mock_events_response.calls[0].response.json()["items"][0] == res.dump(camel_case=True)

    def test_update_multiple(self, cognite_client, mock_events_response):
        res = cognite_client.events.update([EventUpdate(id=1).description.set("blabla")])
        assert isinstance(res, EventList)
        assert mock_events_response.calls[0].response.json()["items"] == res.dump(camel_case=True)

    def test_search(self, cognite_client, mock_events_response):
        res = cognite_client.events.search(filter=EventFilter(external_id_prefix="abc"))
        assert mock_events_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {"search": {"description": None}, "filter": {"externalIdPrefix": "abc"}, "limit": 100} == jsgz_load(
            mock_events_response.calls[0].request.body
        )

    @pytest.mark.parametrize("filter_field", ["external_id_prefix", "externalIdPrefix"])
    def test_search_dict_filter(self, cognite_client, mock_events_response, filter_field):
        res = cognite_client.events.search(filter={filter_field: "bla"})
        assert mock_events_response.calls[0].response.json()["items"] == res.dump(camel_case=True)
        assert {"search": {"description": None}, "filter": {"externalIdPrefix": "bla"}, "limit": 100} == jsgz_load(
            mock_events_response.calls[0].request.body
        )

    def test_event_update_object(self):

        assert isinstance(
            EventUpdate(1)
            .asset_ids.add([])
            .asset_ids.remove([])
            .description.set("")
            .description.set(None)
            .end_time.set(1)
            .end_time.set(None)
            .external_id.set("1")
            .external_id.set(None)
            .metadata.set({})
            .source.set(1)
            .source.set(None)
            .start_time.set(1)
            .start_time.set(None),
            EventUpdate,
        )


@pytest.fixture
def mock_events_empty(rsps, cognite_client):
    url_pattern = re.compile(re.escape(cognite_client.events._get_base_url_with_base_path()) + "/.+")
    rsps.add(rsps.POST, url_pattern, status=200, json={"items": []})
    yield rsps


@pytest.mark.dsl
class TestPandasIntegration:
    def test_event_list_to_pandas(self, cognite_client, mock_events_response):
        import pandas as pd

        df = cognite_client.events.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert 1 == df.shape[0]
        assert {"metadata-key": "metadata-value"} == df["metadata"][0]

    def test_event_list_to_pandas_empty(self, cognite_client, mock_events_empty):
        import pandas as pd

        df = cognite_client.events.list().to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert df.empty

    def test_event_to_pandas(self, cognite_client, mock_events_response):
        import pandas as pd

        df = cognite_client.events.retrieve(id=1).to_pandas()
        assert isinstance(df, pd.DataFrame)
        assert "metadata" not in df.columns
        assert [1] == df.loc["assetIds"][0]
        assert "metadata-value" == df.loc["metadata-key"][0]



import pytest

from cognite.client import CogniteClient
from tests.utils import set_env_var


@pytest.fixture(scope="module")
def cognite_client():
    with set_env_var("COGNITE_API_KEY", "bla"):
        yield CogniteClient()



import gzip
import json as _json
import logging
import numbers
import os
import re
from collections import UserList
from json.decoder import JSONDecodeError
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from urllib.parse import urljoin

import requests.utils
from requests import Response
from requests.structures import CaseInsensitiveDict

from cognite.client import utils
from cognite.client._http_client import GLOBAL_REQUEST_SESSION, HTTPClient, HTTPClientConfig
from cognite.client.data_classes._base import CogniteFilter, CogniteResource, CogniteUpdate
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError

log = logging.getLogger("cognite-sdk")


class APIClient:
    _RESOURCE_PATH = None
    _LIST_CLASS = None

    # TODO: This following set should be generated from the openapi spec somehow.
    RETRYABLE_POST_ENDPOINTS = {
        "/assets/list",
        "/assets/byids",
        "/assets/search",
        "/events/list",
        "/events/byids",
        "/events/search",
        "/files/list",
        "/files/byids",
        "/files/search",
        "/files/downloadlink",
        "/timeseries/byids",
        "/timeseries/search",
        "/timeseries/list",
        "/timeseries/data",
        "/timeseries/data/list",
        "/timeseries/data/latest",
        "/timeseries/data/delete",
        "/sequences/byids",
        "/sequences/search",
        "/sequences/data",
        "/sequences/data/list",
        "/sequences/data/delete",
        "/datasets/list",
        "/datasets/aggregate",
        "/datasets/byids",
        "/relationships/list",
        "/relationships/byids",
        "/context/entitymatching/byids",
        "/context/entitymatching/list",
        "/context/entitymatching/jobs",
    }

    def __init__(self, config: utils._client_config.ClientConfig, api_version: str = None, cognite_client=None) -> None:
        self._config = config
        self._api_version = api_version
        self._api_subversion = config.api_subversion
        self._cognite_client = cognite_client

        session = GLOBAL_REQUEST_SESSION
        if self._config.proxies is not None:
            session.proxies.update(self._config.proxies)

        self._http_client = HTTPClient(
            config=HTTPClientConfig(
                status_codes_to_retry={429},
                backoff_factor=0.5,
                max_backoff_seconds=config.max_retry_backoff,
                max_retries_total=self._config.max_retries,
                max_retries_read=0,
                max_retries_connect=self._config.max_retries,
                max_retries_status=self._config.max_retries,
            ),
            session=session,
        )

        self._http_client_with_retry = HTTPClient(
            config=HTTPClientConfig(
                status_codes_to_retry=self._config.status_forcelist,
                backoff_factor=0.5,
                max_backoff_seconds=config.max_retry_backoff,
                max_retries_total=self._config.max_retries,
                max_retries_read=self._config.max_retries,
                max_retries_connect=self._config.max_retries,
                max_retries_status=self._config.max_retries,
            ),
            session=session,
        )

        self._CREATE_LIMIT = 1000
        self._LIST_LIMIT = 1000
        self._RETRIEVE_LIMIT = 1000
        self._DELETE_LIMIT = 1000
        self._UPDATE_LIMIT = 1000

    def _delete(
        self, url_path: str, params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, Any]] = None
    ) -> requests.Response:
        return self._do_request("DELETE", url_path, params=params, headers=headers, timeout=self._config.timeout)

    def _get(
        self, url_path: str, params: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, Any]] = None
    ) -> requests.Response:
        return self._do_request("GET", url_path, params=params, headers=headers, timeout=self._config.timeout)

    def _post(
        self,
        url_path: str,
        json: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        return self._do_request(
            "POST", url_path, json=json, headers=headers, params=params, timeout=self._config.timeout
        )

    def _put(
        self, url_path: str, json: Optional[Dict[str, Any]] = None, headers: Optional[Dict[str, Any]] = None
    ) -> requests.Response:
        return self._do_request("PUT", url_path, json=json, headers=headers, timeout=self._config.timeout)

    def _do_request(self, method: str, url_path: str, **kwargs) -> requests.Response:
        is_retryable, full_url = self._resolve_url(method, url_path)

        json_payload = kwargs.get("json")
        headers = self._configure_headers(self._config.headers.copy())
        headers.update(kwargs.get("headers") or {})

        if json_payload:
            try:
                data = _json.dumps(json_payload, default=utils._auxiliary.json_dump_default, allow_nan=False)
            except ValueError as e:
                # A lot of work to give a more human friendly error message when nans and infs are present:
                msg = "Out of range float values are not JSON compliant"
                if msg in str(e):  # exc. might e.g. contain an extra ": nan", depending on build (_json.make_encoder)
                    raise ValueError(f"{msg}. Make sure your data does not contain NaN(s) or +/- Inf!").with_traceback(
                        e.__traceback__
                    ) from None
                raise
            kwargs["data"] = data
            if method in ["PUT", "POST"] and not os.getenv("COGNITE_DISABLE_GZIP", False):
                kwargs["data"] = gzip.compress(data.encode())
                headers["Content-Encoding"] = "gzip"

        kwargs["headers"] = headers

        # requests will by default follow redirects. This can be an SSRF-hazard if
        # the client can be tricked to request something with an open redirect, in
        # addition to leaking the token, as requests will send the headers to the
        # redirected-to endpoint.
        # If redirects are to be followed in a call, this should be opted into instead.
        kwargs.setdefault("allow_redirects", False)

        if is_retryable:
            res = self._http_client_with_retry.request(method=method, url=full_url, **kwargs)
        else:
            res = self._http_client.request(method=method, url=full_url, **kwargs)

        if not self._status_ok(res.status_code):
            self._raise_API_error(res, payload=json_payload)
        self._log_request(res, payload=json_payload)
        return res

    def _configure_headers(self, additional_headers: Dict[str, str]) -> CaseInsensitiveDict:
        headers = CaseInsensitiveDict()
        headers.update(requests.utils.default_headers())
        if self._config.token is None:
            headers["api-key"] = self._config.api_key
        elif isinstance(self._config.token, str):
            headers["Authorization"] = "Bearer {}".format(self._config.token)
        elif isinstance(self._config.token, Callable):
            headers["Authorization"] = "Bearer {}".format(self._config.token())
        else:
            raise TypeError("'token' must be str, Callable, or None.")
        headers["content-type"] = "application/json"
        headers["accept"] = "application/json"
        headers["x-cdp-sdk"] = "CognitePythonSDK:{}".format(utils._auxiliary.get_current_sdk_version())
        headers["x-cdp-app"] = self._config.client_name
        headers["cdf-version"] = self._api_subversion
        if "User-Agent" in headers:
            headers["User-Agent"] += " " + utils._auxiliary.get_user_agent()
        else:
            headers["User-Agent"] = utils._auxiliary.get_user_agent()
        headers.update(additional_headers)
        return headers

    def _resolve_url(self, method: str, url_path: str) -> Tuple[bool, str]:
        if not url_path.startswith("/"):
            raise ValueError("URL path must start with '/'")
        base_url = self._get_base_url_with_base_path()
        full_url = base_url + url_path
        is_retryable = self._is_retryable(method, full_url)
        return is_retryable, full_url

    def _get_base_url_with_base_path(self) -> str:
        base_path = "/api/{}/projects/{}".format(self._api_version, self._config.project) if self._api_version else ""
        return urljoin(self._config.base_url, base_path)

    def _is_retryable(self, method: str, path: str) -> bool:
        valid_methods = ["GET", "POST", "PUT", "DELETE", "PATCH"]
        match = re.match(
            "(?:http|https)://[a-z\d.:\-]+(?:/api/(?:v1|playground)/projects/[^/]+)?(/[^\?]+)?(\?.+)?", path
        )
        if not match:
            raise ValueError("Path {} is not valid. Cannot resolve whether or not it is retryable".format(path))
        if method not in valid_methods:
            raise ValueError("Method {} is not valid. Must be one of {}".format(method, valid_methods))
        path_end = match.group(1)

        if method in ["GET", "PUT", "PATCH"]:
            return True
        if method == "POST" and path_end in self.RETRYABLE_POST_ENDPOINTS:
            return True
        return False

    def _retrieve(
        self, id: Union[int, str], cls=None, resource_path: str = None, params: Dict = None, headers: Dict = None
    ):
        cls = cls or self._LIST_CLASS._RESOURCE
        resource_path = resource_path or self._RESOURCE_PATH
        try:
            res = self._get(
                url_path=utils._auxiliary.interpolate_and_url_encode(resource_path + "/{}", str(id)),
                params=params,
                headers=headers,
            )
            return cls._load(res.json(), cognite_client=self._cognite_client)
        except CogniteAPIError as e:
            if e.code != 404:
                raise

    def _retrieve_multiple(
        self,
        wrap_ids: bool,
        cls=None,
        resource_path: str = None,
        ids: Union[List[int], int] = None,
        external_ids: Union[List[str], str] = None,
        ignore_unknown_ids=None,
        headers: Dict = None,
        other_params: Dict = None,
    ):
        cls = cls or self._LIST_CLASS
        resource_path = resource_path or self._RESOURCE_PATH
        all_ids = self._process_ids(ids, external_ids, wrap_ids=wrap_ids)
        id_chunks = utils._auxiliary.split_into_chunks(all_ids, self._RETRIEVE_LIMIT)

        ignore_unknown = {} if ignore_unknown_ids is None else {"ignoreUnknownIds": ignore_unknown_ids}
        tasks = [
            {
                "url_path": resource_path + "/byids",
                "json": {"items": id_chunk, **ignore_unknown, **(other_params or {})},
                "headers": headers,
            }
            for id_chunk in id_chunks
        ]
        tasks_summary = utils._concurrency.execute_tasks_concurrently(
            self._post, tasks, max_workers=self._config.max_workers
        )

        if tasks_summary.exceptions:
            try:
                utils._concurrency.collect_exc_info_and_raise(tasks_summary.exceptions)
            except CogniteNotFoundError:
                if self._is_single_identifier(ids, external_ids):
                    return None
                raise

        retrieved_items = tasks_summary.joined_results(lambda res: res.json()["items"])

        if self._is_single_identifier(ids, external_ids):
            return cls._RESOURCE._load(retrieved_items[0], cognite_client=self._cognite_client)
        return cls._load(retrieved_items, cognite_client=self._cognite_client)

    def _list_generator(
        self,
        method: str,
        cls=None,
        resource_path: str = None,
        url_path: str = None,
        limit: int = None,
        chunk_size: int = None,
        filter: Dict = None,
        sort: List[str] = None,
        other_params: Dict = None,
        partitions: int = None,
        headers: Dict = None,
        initial_cursor: Optional[str] = None,
    ):
        if limit == -1 or limit == float("inf"):
            limit = None
        cls = cls or self._LIST_CLASS
        resource_path = resource_path or self._RESOURCE_PATH

        if partitions:
            if limit is not None:
                raise ValueError("When using partitions, limit should be `None`, `-1` or `inf`.")
            if sort is not None:
                raise ValueError("When using sort, partitions is not supported.")

            yield from self._list_generator_partitioned(
                partitions=partitions,
                cls=cls,
                resource_path=resource_path,
                filter=filter,
                other_params=other_params,
                headers=headers,
            )

        else:
            total_items_retrieved = 0
            current_limit = self._LIST_LIMIT
            if chunk_size and chunk_size <= self._LIST_LIMIT:
                current_limit = chunk_size
            next_cursor = initial_cursor
            filter = filter or {}
            current_items = []
            while True:
                if limit:
                    num_of_remaining_items = limit - total_items_retrieved
                    if num_of_remaining_items < self._LIST_LIMIT:
                        current_limit = num_of_remaining_items

                if method == "GET":
                    params = filter.copy()
                    params["limit"] = current_limit
                    params["cursor"] = next_cursor
                    if sort is not None:
                        params["sort"] = sort
                    res = self._get(url_path=url_path or resource_path, params=params, headers=headers)
                elif method == "POST":
                    body = {"filter": filter, "limit": current_limit, "cursor": next_cursor, **(other_params or {})}
                    if sort is not None:
                        body["sort"] = sort
                    res = self._post(url_path=url_path or resource_path + "/list", json=body, headers=headers)
                else:
                    raise ValueError("_list_generator parameter `method` must be GET or POST, not {}".format(method))
                last_received_items = res.json()["items"]
                total_items_retrieved += len(last_received_items)

                if not chunk_size:
                    for item in last_received_items:
                        yield cls._RESOURCE._load(item, cognite_client=self._cognite_client)
                else:
                    current_items.extend(last_received_items)
                    if len(current_items) >= chunk_size:
                        items_to_yield = current_items[:chunk_size]
                        current_items = current_items[chunk_size:]
                        yield cls._load(items_to_yield, cognite_client=self._cognite_client)

                next_cursor = res.json().get("nextCursor")
                if total_items_retrieved == limit or next_cursor is None:
                    if chunk_size and current_items:
                        yield cls._load(current_items, cognite_client=self._cognite_client)
                    break

    def _list_generator_partitioned(
        self,
        partitions: int,
        cls,
        resource_path: str,
        filter: Dict = None,
        other_params: Dict = None,
        headers: Dict = None,
    ):
        next_cursors = {i + 1: None for i in range(partitions)}

        def get_partition(partition_num):
            next_cursor = next_cursors[partition_num]

            body = {
                "filter": filter or {},
                "limit": self._LIST_LIMIT,
                "cursor": next_cursor,
                "partition": "{}/{}".format(partition_num, partitions),
                **(other_params or {}),
            }
            res = self._post(url_path=resource_path + "/list", json=body, headers=headers)
            next_cursor = res.json().get("nextCursor")
            next_cursors[partition_num] = next_cursor

            return res.json()["items"]

        while len(next_cursors) > 0:
            tasks_summary = utils._concurrency.execute_tasks_concurrently(
                get_partition, [(partition,) for partition in next_cursors], max_workers=partitions
            )
            if tasks_summary.exceptions:
                raise tasks_summary.exceptions[0]

            for item in tasks_summary.joined_results():
                yield cls._RESOURCE._load(item, cognite_client=self._cognite_client)

            # Remove None from cursor dict
            next_cursors = {partition: next_cursors[partition] for partition in next_cursors if next_cursors[partition]}

    def _list(
        self,
        method: str,
        cls=None,
        resource_path: str = None,
        url_path: str = None,
        limit: int = None,
        filter: Dict = None,
        other_params=None,
        partitions=None,
        sort=None,
        headers: Dict = None,
        initial_cursor: Optional[str] = None,
    ):
        if partitions:
            if limit not in [None, -1, float("inf")]:
                raise ValueError("When using partitions, limit should be `None`, `-1` or `inf`.")
            if sort is not None:
                raise ValueError("When using sort, partitions is not supported.")
            return self._list_partitioned(
                partitions=partitions,
                method=method,
                cls=cls,
                resource_path=resource_path,
                filter=filter,
                other_params=other_params,
                headers=headers,
            )

        cls = cls or self._LIST_CLASS
        resource_path = resource_path or self._RESOURCE_PATH
        items = []
        for resource_list in self._list_generator(
            cls=cls,
            resource_path=resource_path,
            url_path=url_path,
            method=method,
            limit=limit,
            chunk_size=self._LIST_LIMIT,
            filter=filter,
            sort=sort,
            other_params=other_params,
            headers=headers,
            initial_cursor=initial_cursor,
        ):
            items.extend(resource_list.data)
        return cls(items, cognite_client=self._cognite_client)

    def _list_partitioned(
        self,
        partitions,
        method,
        cls=None,
        resource_path: str = None,
        filter: Dict = None,
        other_params=None,
        headers: Dict = None,
    ):
        cls = cls or self._LIST_CLASS
        resource_path = resource_path or self._RESOURCE_PATH

        def get_partition(partition):
            next_cursor = None
            retrieved_items = []
            while True:
                if method == "POST":
                    body = {
                        "filter": filter or {},
                        "limit": self._LIST_LIMIT,
                        "cursor": next_cursor,
                        "partition": partition,
                        **(other_params or {}),
                    }
                    res = self._post(url_path=resource_path + "/list", json=body, headers=headers)
                elif method == "GET":
                    params = {
                        **(filter or {}),
                        "limit": self._LIST_LIMIT,
                        "cursor": next_cursor,
                        "partition": partition,
                        **(other_params or {}),
                    }
                    res = self._get(url_path=resource_path, params=params, headers=headers)
                retrieved_items.extend(res.json()["items"])
                next_cursor = res.json().get("nextCursor")
                if next_cursor is None:
                    break
            return retrieved_items

        tasks = [("{}/{}".format(i + 1, partitions),) for i in range(partitions)]
        tasks_summary = utils._concurrency.execute_tasks_concurrently(get_partition, tasks, max_workers=partitions)
        if tasks_summary.exceptions:
            raise tasks_summary.exceptions[0]
        return cls._load(tasks_summary.joined_results(), cognite_client=self._cognite_client)

    def _aggregate(
        self,
        resource_path: str = None,
        filter: Union[CogniteFilter, Dict] = None,
        aggregate: str = None,
        fields: List[str] = None,
        headers: Dict = None,
        cls=None,
    ):
        utils._auxiliary.assert_type(filter, "filter", [dict, CogniteFilter], allow_none=True)
        utils._auxiliary.assert_type(fields, "fields", [list], allow_none=True)
        if isinstance(filter, CogniteFilter):
            filter = filter.dump(camel_case=True)
        elif isinstance(filter, Dict):
            filter = utils._auxiliary.convert_all_keys_to_camel_case(filter)
        resource_path = resource_path or self._RESOURCE_PATH
        body = {"filter": filter or {}}
        if aggregate is not None:
            body["aggregate"] = aggregate
        if fields is not None:
            body["fields"] = fields
        res = self._post(url_path=resource_path + "/aggregate", json=body, headers=headers)
        return [cls(**agg) for agg in res.json()["items"]]

    def _create_multiple(
        self,
        items: Union[List[Any], Any],
        cls: Any = None,
        resource_path: str = None,
        params: Dict = None,
        headers: Dict = None,
        limit=None,
    ):
        cls = cls or self._LIST_CLASS
        resource_path = resource_path or self._RESOURCE_PATH
        limit = limit or self._CREATE_LIMIT
        single_item = not isinstance(items, list)
        if single_item:
            items = [items]

        items_split = []
        for i in range(0, len(items), limit):
            if isinstance(items[i], CogniteResource):
                items_chunk = [item.dump(camel_case=True) for item in items[i : i + limit]]
            else:
                items_chunk = [item for item in items[i : i + limit]]
            items_split.append({"items": items_chunk})

        tasks = [(resource_path, task_items, params, headers) for task_items in items_split]
        summary = utils._concurrency.execute_tasks_concurrently(self._post, tasks, max_workers=self._config.max_workers)

        def unwrap_element(el):
            if isinstance(el, dict):
                return cls._RESOURCE._load(el, cognite_client=self._cognite_client)
            else:
                return el

        def str_format_element(el):
            if isinstance(el, CogniteResource):
                dumped = el.dump()
                if "external_id" in dumped:
                    return dumped["external_id"]
                return dumped
            return el

        summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=lambda task: task[1]["items"],
            task_list_element_unwrap_fn=unwrap_element,
            str_format_element_fn=str_format_element,
        )
        created_resources = summary.joined_results(lambda res: res.json()["items"])

        if single_item:
            return cls._RESOURCE._load(created_resources[0], cognite_client=self._cognite_client)
        return cls._load(created_resources, cognite_client=self._cognite_client)

    def _delete_multiple(
        self,
        wrap_ids: bool,
        resource_path: str = None,
        ids: Union[List[int], int] = None,
        external_ids: Union[List[str], str] = None,
        params: Dict = None,
        headers: Dict = None,
        extra_body_fields: Dict = None,
    ):
        resource_path = resource_path or self._RESOURCE_PATH
        all_ids = self._process_ids(ids, external_ids, wrap_ids)
        id_chunks = utils._auxiliary.split_into_chunks(all_ids, self._DELETE_LIMIT)
        tasks = [
            {
                "url_path": resource_path + "/delete",
                "json": {"items": chunk, **(extra_body_fields or {})},
                "params": params,
                "headers": headers,
            }
            for chunk in id_chunks
        ]
        summary = utils._concurrency.execute_tasks_concurrently(self._post, tasks, max_workers=self._config.max_workers)
        summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=lambda task: task["json"]["items"],
            task_list_element_unwrap_fn=utils._auxiliary.unwrap_identifer,
        )

    def _update_multiple(
        self,
        items: Union[List[Any], Any],
        cls: Any = None,
        resource_path: str = None,
        params: Dict = None,
        headers: Dict = None,
    ):
        cls = cls or self._LIST_CLASS
        resource_path = resource_path or self._RESOURCE_PATH
        patch_objects = []
        single_item = not isinstance(items, (list, UserList))
        if single_item:
            items = [items]

        for index, item in enumerate(items):
            if isinstance(item, CogniteResource):
                patch_objects.append(self._convert_resource_to_patch_object(item, cls._UPDATE._get_update_properties()))
            elif isinstance(item, CogniteUpdate):
                patch_objects.append(item.dump())
                patch_object_update = patch_objects[index]["update"]
                if "metadata" in patch_object_update and patch_object_update["metadata"] == {"set": None}:
                    patch_object_update["metadata"] = {"set": {}}
            else:
                raise ValueError("update item must be of type CogniteResource or CogniteUpdate")
        patch_object_chunks = utils._auxiliary.split_into_chunks(patch_objects, self._UPDATE_LIMIT)

        tasks = [
            {"url_path": resource_path + "/update", "json": {"items": chunk}, "params": params, "headers": headers}
            for chunk in patch_object_chunks
        ]

        tasks_summary = utils._concurrency.execute_tasks_concurrently(
            self._post, tasks, max_workers=self._config.max_workers
        )
        tasks_summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=lambda task: task["json"]["items"],
            task_list_element_unwrap_fn=lambda el: utils._auxiliary.unwrap_identifer(el),
        )
        updated_items = tasks_summary.joined_results(lambda res: res.json()["items"])

        if single_item:
            return cls._RESOURCE._load(updated_items[0], cognite_client=self._cognite_client)
        return cls._load(updated_items, cognite_client=self._cognite_client)

    def _search(
        self,
        search: Dict,
        filter: Union[Dict, CogniteFilter],
        limit: int,
        cls: Any = None,
        resource_path: str = None,
        params: Dict = None,
        headers: Dict = None,
    ):
        utils._auxiliary.assert_type(filter, "filter", [dict, CogniteFilter], allow_none=True)
        if isinstance(filter, CogniteFilter):
            filter = filter.dump(camel_case=True)
        elif isinstance(filter, dict):
            filter = utils._auxiliary.convert_all_keys_to_camel_case(filter)
        cls = cls or self._LIST_CLASS
        resource_path = resource_path or self._RESOURCE_PATH
        res = self._post(
            url_path=resource_path + "/search",
            json={"search": search, "filter": filter, "limit": limit},
            params=params,
            headers=headers,
        )
        return cls._load(res.json()["items"], cognite_client=self._cognite_client)

    @staticmethod
    def _convert_resource_to_patch_object(resource, update_attributes):
        dumped_resource = resource.dump(camel_case=True)
        has_id = "id" in dumped_resource
        has_external_id = "externalId" in dumped_resource

        patch_object = {"update": {}}
        if has_id:
            patch_object["id"] = dumped_resource.pop("id")
        elif has_external_id:
            patch_object["externalId"] = dumped_resource.pop("externalId")

        for key, value in dumped_resource.items():
            if utils._auxiliary.to_snake_case(key) in update_attributes:
                patch_object["update"][key] = {"set": value}
        return patch_object

    @staticmethod
    def _process_ids(
        ids: Union[List[int], int, None], external_ids: Union[List[str], str, None], wrap_ids: bool
    ) -> List:
        if external_ids is None and ids is None:
            raise ValueError("No ids specified")
        if external_ids and not wrap_ids:
            raise ValueError("externalIds must be wrapped")

        if isinstance(ids, numbers.Integral):
            ids = [ids]
        elif isinstance(ids, list) or ids is None:
            ids = ids or []
        else:
            raise TypeError("ids must be int or list of int")

        if isinstance(external_ids, str):
            external_ids = [external_ids]
        elif isinstance(external_ids, list) or external_ids is None:
            external_ids = external_ids or []
        else:
            raise TypeError("external_ids must be str or list of str")

        if wrap_ids:
            ids = [{"id": id} for id in ids]
            external_ids = [{"externalId": external_id} for external_id in external_ids]

        all_ids = ids + external_ids

        return all_ids

    @staticmethod
    def _is_single_identifier(ids, external_ids):
        single_id = isinstance(ids, numbers.Integral) and external_ids is None
        single_external_id = isinstance(external_ids, str) and ids is None
        return single_id or single_external_id

    @staticmethod
    def _status_ok(status_code: int):
        return status_code in {200, 201, 202}

    @classmethod
    def _raise_API_error(cls, res: Response, payload: Dict):
        x_request_id = res.headers.get("X-Request-Id")
        code = res.status_code
        missing = None
        duplicated = None
        extra = {}
        try:
            error = res.json()["error"]
            if isinstance(error, str):
                msg = error
            elif isinstance(error, Dict):
                msg = error["message"]
                missing = error.get("missing")
                duplicated = error.get("duplicated")
                for k, v in error.items():
                    if k not in ["message", "missing", "duplicated", "code"]:
                        extra[k] = v
            else:
                msg = res.content
        except Exception:
            msg = res.content

        error_details = {"X-Request-ID": x_request_id}
        if payload:
            error_details["payload"] = payload
        if missing:
            error_details["missing"] = missing
        if duplicated:
            error_details["duplicated"] = duplicated
        error_details["headers"] = res.request.headers.copy()
        APIClient._sanitize_headers(error_details["headers"])
        error_details["response_payload"] = cls._truncate(cls._get_response_content_safe(res))
        error_details["response_headers"] = res.headers

        if res.history:
            for res_hist in res.history:
                log.debug(
                    "REDIRECT AFTER HTTP Error {} {} {}: {}".format(
                        res_hist.status_code, res_hist.request.method, res_hist.request.url, res_hist.content
                    )
                )
        log.debug("HTTP Error {} {} {}: {}".format(code, res.request.method, res.request.url, msg), extra=error_details)
        raise CogniteAPIError(msg, code, x_request_id, missing=missing, duplicated=duplicated, extra=extra)

    @classmethod
    def _log_request(cls, res: Response, **kwargs):
        method = res.request.method
        url = res.request.url
        status_code = res.status_code

        extra = kwargs.copy()
        extra["headers"] = res.request.headers.copy()
        APIClient._sanitize_headers(extra["headers"])
        if extra["payload"] is None:
            del extra["payload"]

        extra["response_payload"] = cls._truncate(cls._get_response_content_safe(res))
        extra["response_headers"] = res.headers

        http_protocol_version = ".".join(list(str(res.raw.version)))

        log.debug("HTTP/{} {} {} {}".format(http_protocol_version, method, url, status_code), extra=extra)

    @staticmethod
    def _truncate(s: str, limit: int = 500) -> str:
        if len(s) > limit:
            return s[:limit] + "..."
        return s

    @classmethod
    def _get_response_content_safe(cls, res: Response) -> str:
        try:
            return _json.dumps(res.json())
        except JSONDecodeError:
            pass

        try:
            return res.content.decode()
        except UnicodeDecodeError:
            pass

        return "<binary>"

    @staticmethod
    def _sanitize_headers(headers: Optional[Dict]):
        if headers is None:
            return
        if "api-key" in headers:
            headers["api-key"] = "***"
        if "Authorization" in headers:
            headers["Authorization"] = "***"



import asyncio
import random
import string
import time

import pytest

from cognite.client.data_classes import (
    OidcCredentials,
    Transformation,
    TransformationDestination,
    TransformationJobStatus,
)


@pytest.fixture
def new_transformation(cognite_client):
    prefix = "".join(random.choice(string.ascii_letters) for i in range(6))
    transform = Transformation(
        name="any",
        external_id=f"{prefix}-transformation",
        destination=TransformationDestination.assets(),
        query="select id, name from _cdf.assets limit 1000",
        source_oidc_credentials=OidcCredentials(
            client_id=cognite_client.config.token_client_id,
            client_secret=cognite_client.config.token_client_secret,
            scopes=",".join(cognite_client.config.token_scopes),
            token_uri=cognite_client.config.token_url,
            cdf_project_name=cognite_client.config.project,
        ),
        destination_oidc_credentials=OidcCredentials(
            client_id=cognite_client.config.token_client_id,
            client_secret=cognite_client.config.token_client_secret,
            scopes=",".join(cognite_client.config.token_scopes),
            token_uri=cognite_client.config.token_url,
            cdf_project_name=cognite_client.config.project,
        ),
        ignore_null_fields=True,
    )
    ts = cognite_client.transformations.create(transform)

    yield ts

    cognite_client.transformations.delete(id=ts.id)
    assert cognite_client.transformations.retrieve(ts.id) is None


@pytest.fixture
def new_raw_transformation(cognite_client, new_transformation):
    new_transformation.query = "select 1 as key, 'example2' as name, 'example' as externalId"
    new_transformation.destination = TransformationDestination.raw("my_db", "my_table")
    ts = cognite_client.transformations.update(new_transformation)

    yield ts


other_transformation = new_transformation


@pytest.fixture
def longer_transformation(cognite_client, new_transformation):
    new_transformation.query = "select id, name from _cdf.assets limit 5000"
    ts = cognite_client.transformations.update(new_transformation)

    yield ts


async def run_transformation_without_waiting(new_transformation):
    job = new_transformation.run(wait=False)

    yield (job, new_transformation)

    await job.wait_async()
    assert job.status == TransformationJobStatus.COMPLETED


@pytest.fixture
async def new_running_transformation(new_transformation):
    async for transform in run_transformation_without_waiting(new_transformation):
        yield transform


@pytest.fixture
async def other_running_transformation(other_transformation):
    async for transform in run_transformation_without_waiting(other_transformation):
        yield transform


class TestTransformationJobsAPI:
    @pytest.mark.asyncio
    async def test_run_without_wait(self, cognite_client, new_running_transformation):
        (job, new_transformation) = new_running_transformation
        assert (
            job.id is not None
            and job.status == TransformationJobStatus.CREATED
            and job.transformation_id == new_transformation.id
            and job.transformation_external_id == new_transformation.external_id
            and job.source_project == cognite_client.config.project
            and job.destination_project == cognite_client.config.project
            and job.destination == TransformationDestination.assets()
            and job.conflict_mode == "upsert"
            and job.query == new_transformation.query
            and job.error is None
            and job.ignore_null_fields
        )
        await asyncio.sleep(0.5)
        retrieved_transformation = cognite_client.transformations.retrieve(id=new_transformation.id)

        assert retrieved_transformation.running_job is not None and retrieved_transformation.running_job.id == job.id

    def test_run(self, cognite_client, new_transformation: Transformation):
        job = new_transformation.run()

        assert (
            job.id is not None
            and job.status == TransformationJobStatus.COMPLETED
            and job.transformation_id == new_transformation.id
            and job.transformation_external_id == new_transformation.external_id
            and job.source_project == cognite_client.config.project
            and job.destination_project == cognite_client.config.project
            and job.destination == TransformationDestination.assets()
            and job.conflict_mode == "upsert"
            and job.query == new_transformation.query
            and job.error is None
            and job.ignore_null_fields
        )

        retrieved_transformation = cognite_client.transformations.retrieve(id=new_transformation.id)

        assert (
            retrieved_transformation.last_finished_job is not None
            and retrieved_transformation.last_finished_job.id == job.id
        )

    def test_run_with_timeout(self, longer_transformation: Transformation):
        init = time.time()
        timeout = 0.1
        job = longer_transformation.run(timeout=timeout)
        final = time.time()

        assert job.status == TransformationJobStatus.RUNNING and timeout <= final - init <= timeout + 1.5

    @pytest.mark.asyncio
    async def test_run_async(self, cognite_client, new_transformation: Transformation):
        job = await new_transformation.run_async()

        assert (
            job.id is not None
            and job.status == TransformationJobStatus.COMPLETED
            and job.transformation_id == new_transformation.id
            and job.transformation_external_id == new_transformation.external_id
            and job.source_project == cognite_client.config.project
            and job.destination_project == cognite_client.config.project
            and job.destination == TransformationDestination.assets()
            and job.conflict_mode == "upsert"
            and job.query == new_transformation.query
            and job.error is None
            and job.ignore_null_fields
        )

    @pytest.mark.asyncio
    async def test_run_with_timeout_async(self, longer_transformation: Transformation):
        init = time.time()
        timeout = 0.1
        job = await longer_transformation.run_async(timeout=timeout)
        final = time.time()

        assert job.status == TransformationJobStatus.RUNNING and timeout <= final - init <= timeout + 1.5

    @pytest.mark.asyncio
    async def test_run_by_external_id_async(self, cognite_client, new_transformation: Transformation):
        job = await cognite_client.transformations.run_async(transformation_external_id=new_transformation.external_id)

        assert (
            job.id is not None
            and job.status == TransformationJobStatus.COMPLETED
            and job.transformation_id == new_transformation.id
            and job.transformation_external_id == new_transformation.external_id
            and job.source_project == cognite_client.config.project
            and job.destination_project == cognite_client.config.project
            and job.destination == TransformationDestination.assets()
            and job.conflict_mode == "upsert"
            and job.query == new_transformation.query
            and job.error is None
            and job.ignore_null_fields
        )

    @pytest.mark.asyncio
    async def test_run_raw_transformation(self, cognite_client, new_raw_transformation):
        job = await new_raw_transformation.run_async()

        assert (
            job.id is not None
            and job.status not in [TransformationJobStatus.CREATED, TransformationJobStatus.RUNNING]
            and job.transformation_id == new_raw_transformation.id
            and job.transformation_external_id == new_raw_transformation.external_id
            and job.source_project == cognite_client.config.project
            and job.destination_project == cognite_client.config.project
            and job.destination == TransformationDestination.raw("my_db", "my_table")
            and job.conflict_mode == "upsert"
            and job.query == new_raw_transformation.query
            and job.ignore_null_fields
        )

    @pytest.mark.asyncio
    async def test_list_jobs_by_transformation_id(self, new_running_transformation):
        (new_job, new_transformation) = new_running_transformation

        retrieved_jobs = new_transformation.jobs()
        assert new_job.id in [job.id for job in retrieved_jobs]
        assert all(job.transformation_id == new_transformation.id for job in retrieved_jobs)

    @pytest.mark.asyncio
    async def test_list_jobs(self, cognite_client, new_running_transformation, other_running_transformation):
        (new_job, _) = new_running_transformation
        (other_job, _) = other_running_transformation

        retrieved_jobs = cognite_client.transformations.jobs.list()
        assert new_job.id in [job.id for job in retrieved_jobs]
        assert other_job.id in [job.id for job in retrieved_jobs]

    @pytest.mark.asyncio
    async def test_metrics(self, new_running_transformation):
        (job, _) = new_running_transformation
        await asyncio.sleep(1.0)
        metrics = job.metrics()
        assert metrics is not None

    @pytest.mark.asyncio
    async def test_retrieve_multiple(self, cognite_client, new_running_transformation, other_running_transformation):
        (new_job, _) = new_running_transformation
        (other_job, _) = other_running_transformation

        retrieved_jobs = cognite_client.transformations.jobs.retrieve_multiple(ids=[new_job.id, other_job.id])
        assert new_job.id in [job.id for job in retrieved_jobs]
        assert other_job.id in [job.id for job in retrieved_jobs]
        assert len(retrieved_jobs) == 2



import pytest

from cognite.client.data_classes import *


@pytest.mark.dsl
class TestRepr:
    def test_repr_html(self):
        for cls in [Asset, Datapoints, Sequence, FileMetadata, Row, Table, ThreeDModel]:
            assert len(cls()._repr_html_()) > 0

    def test_repr_html_datapoint(self):
        assert len(Datapoint(timestamp=0, value=0)._repr_html_()) > 0

    def test_repr_html_list(self):
        for cls in [AssetList, DatapointsList, RowList, TableList]:
            assert len(cls([cls._RESOURCE()])._repr_html_()) > 0



import json
import re
from random import random

import pytest

from cognite.client.data_classes import Datapoints
from tests.utils import jsgz_load


def generate_datapoints(start: int, end: int, granularity=1):
    dps = []
    for i in range(start, end, granularity):
        dp = {}
        dp["value"] = random()
        dp["timestamp"] = i
        dps.append(dp)
    return dps


@pytest.fixture
def mock_get_datapoints(rsps, cognite_client):
    def request_callback(request):
        payload = jsgz_load(request.body)

        items = []
        for dps_query in payload["items"]:

            if "start" in dps_query and "end" in dps_query:
                start, end = dps_query["start"], dps_query["end"]
            else:
                start, end = payload["start"], payload["end"]

            limit = 10000
            if "limit" in dps_query:
                limit = dps_query["limit"]
            elif "limit" in payload:
                limit = payload["limit"]

            dps = generate_datapoints(start, end)
            dps = dps[:limit]
            items.append({"isString": False, "datapoints": dps})
        response = {"items": items}
        return 200, {}, json.dumps(response)

    rsps.add_callback(
        rsps.POST,
        cognite_client.datapoints.synthetic._get_base_url_with_base_path() + "/timeseries/synthetic/query",
        callback=request_callback,
        content_type="application/json",
    )
    yield rsps


@pytest.fixture
def mock_get_datapoints_empty(rsps, cognite_client):
    rsps.add(
        rsps.POST,
        re.compile(
            re.escape(cognite_client.datapoints.synthetic._get_base_url_with_base_path()) + "/timeseries/synthetic/.*"
        ),
        status=200,
        json={"items": [{"isString": False, "datapoints": []}]},
    )
    yield rsps


class TestSyntheticQuery:
    def test_query(self, cognite_client, mock_get_datapoints):
        dps_res = cognite_client.datapoints.synthetic.query(
            expressions='TS{externalID:"abc"} + TS{id:1}', start=1000000, end=1100001
        )
        assert isinstance(dps_res, Datapoints)
        assert 100001 == len(dps_res)
        assert 11 == len(mock_get_datapoints.calls)

    def test_query_limit(self, cognite_client, mock_get_datapoints):
        dps_res = cognite_client.datapoints.synthetic.query(
            expressions=['TS{externalID:"abc"}', "TS{id:1}"], start=1000000, end=1100001, limit=20000
        )
        assert 20000 == len(dps_res[0])
        assert 20000 == len(dps_res[1])
        assert 4 == len(mock_get_datapoints.calls)

    def test_query_empty(self, cognite_client, mock_get_datapoints_empty):
        dps_res = cognite_client.datapoints.synthetic.query(
            expressions=['TS{externalID:"abc"} + TS{id:1}'], start=1000000, end=1100001
        )
        assert isinstance(dps_res[0], Datapoints)
        assert 0 == len(dps_res[0])
        assert 1 == len(mock_get_datapoints_empty.calls)

    @pytest.mark.dsl
    def test_expression_builder(self, cognite_client):
        from sympy import symbols

        assert ("ts{externalId:'x'}", "a") == cognite_client.datapoints.synthetic._build_expression(
            symbols("a"), {"a": "x"}
        )
        assert (
            "ts{externalId:'x',aggregate:'average',granularity:'1m'}",
            "a",
        ) == cognite_client.datapoints.synthetic._build_expression(
            symbols("a"), {"a": "x"}, aggregate="average", granularity="1m"
        )
        assert (
            "(ts{externalId:'x'}+ts{externalId:'y'}+ts{externalId:'z'})",
            "(a+b+c)",
        ) == cognite_client.datapoints.synthetic._build_expression(
            symbols("a") + symbols("b") + symbols("c"), {"a": "x", "b": "y", "c": "z"}
        )
        assert ("(1/ts{externalId:'a'})", "(1/a)") == cognite_client.datapoints.synthetic._build_expression(
            1 / symbols("a"), {"a": "a"}
        )

    @pytest.mark.dsl
    def test_expression_builder_variables_missing(self, cognite_client):
        from sympy import symbols

        with pytest.raises(
            ValueError, match="sympy expressions are only supported in combination with the `variables` parameter"
        ):
            cognite_client.datapoints.synthetic.query([symbols("a")], start=0, end="now")

    @pytest.mark.dsl
    def test_expression_builder_unsupported_missing(self, cognite_client):
        from sympy import cot, symbols

        with pytest.raises(ValueError, match="Unsupported sympy class cot"):
            cognite_client.datapoints.synthetic.query(
                [symbols("a") + cot(symbols("a"))], start=0, end="now", variables={"a": "a"}
            )
